''' Utilities for dealing with fastq files. '''

import gzip
import functools
import array
from pathlib import Path
from itertools import chain
from collections import namedtuple

import numpy as np

from . import utilities
from .fastq_cython import *
from .utilities import identity, base_order, group_by

# SANGER_OFFSET = 33 is imported from fastq_cython
SOLEXA_OFFSET = 64
MAX_QUAL = 93
MAX_EXPECTED_QUAL = 42

def decode_sanger(qual):
    ''' Converts a string of sanger-encoded quals to a list of integers. '''
    return [ord(q) - SANGER_OFFSET for q in qual]

def decode_sanger_to_array(qual):
    ''' Converts a string of sanger-encoded quals to a (python, not numpy) array of unsigned chars
    suitable for assignment to a pysam.AlignedSegment.query_qualities '''
    q_array = array.array('B', qual.encode())
    subtract_sanger_offset(q_array)
    return q_array

def decode_solexa(qual):
    ''' Converts a string of solexa-encoded quals to a list of integers. '''
    return [ord(q) - SOLEXA_OFFSET for q in qual]

def encode_sanger(ints):
    ''' Converts a list of integer quals to a sanger-encoded string. '''
    return ''.join(chr(i + SANGER_OFFSET) for i in ints)

def encode_solexa(ints):
    ''' Converts a list of integer quals to a solexa-encoded string. '''
    return ''.join(chr(i + SOLEXA_OFFSET) for i in ints)

_solexa_to_sanger_table = {}
for q in range(ord('!') - SOLEXA_OFFSET, MAX_EXPECTED_QUAL + 1):
    # Old solexa encoding was -10 log(p / (1 - p)), which could be negative.
    # Character encodings of negative values cause problems, and we don't really
    # care about fine distinctions in low quality scores, so just set to a
    # minimum of zero.
    # Some files assign ! to N's, so this range needs to go down to ord('!').
    _solexa_to_sanger_table[chr(q + SOLEXA_OFFSET)] = chr(max(q, 0) + SANGER_OFFSET)

def solexa_to_sanger(qual):
    return ''.join(_solexa_to_sanger_table[c] for c in qual)

# If a qname for a paired mapping ends in '/1', '/2', or '/3', bowtie2 chops off
# the last two characters of the qname. If qual strings of trimmed portions of
# reads are to be put in qnames, '/' needs to be downgraded to
# chr(ord('/') - 1). 
# '_' is used as a field separator in annotations, so similarly needs to be
# downgraded to chr(ord('_') - 1).
_chars_to_sanitize = '/_'
_sanitized_chars = ''.join(chr(ord(c) - 1) for c in _chars_to_sanitize)
try:
    _sanitize_table = str.maketrans(_chars_to_sanitize, _sanitized_chars)
    period_to_N = str.maketrans('.', 'N')
except AttributeError:
    import string
    _sanitize_table = string.maketrans(_chars_to_sanitize, _sanitized_chars)
    period_to_N = string.maketrans('.', 'N')

def sanitize_qual(qual):
    sanitized = qual.translate(_sanitize_table)
    return sanitized

def quality_and_complexity(reads_iter, max_read_length, alignments=False, min_q=0):
    stats = {
        'q': np.zeros((max_read_length, MAX_EXPECTED_QUAL + 1), int),
        'c': np.zeros((max_read_length, 256), int),
        'c_above_min_q': np.zeros((max_read_length, 256), int),
        'average_q': np.zeros((max_read_length, 256), int),
    }

    if isinstance(reads_iter, (str, Path)):
        reads_iter = reads(reads_iter)

    for read in reads_iter:
        if alignments:
            process_Alignment(read.query_sequence.encode(),
                              read.query_qualities,
                              stats['q'],
                              stats['average_q'],
                              stats['c'],
                              stats['c_above_min_q'],
                              min_q,
                             )
        else:
            process_read(read.seq.encode(), read.qual.encode(),
                         stats['q'],
                         stats['average_q'],
                         stats['c'],
                         stats['c_above_min_q'],
                         min_q,
                        )
        
    # To avoid a lookup at every single base, base-specific arrays are 2*max_read_length x 256.
    # This pulls out only the columns corresponding to possible base
    # identities. 
    for k in ['c', 'c_above_min_q', 'average_q']:
        stats[k] = np.vstack([stats[k].T[ord(b)] for b in base_order]).T

    stats['average_q'] = stats['average_q'] / np.maximum(1, stats['c'])
    
    return stats

def quality_and_complexity_paired(read_pairs, max_read_length):
    R1_q_array = np.zeros((max_read_length, MAX_EXPECTED_QUAL + 1), int)
    R1_c_array = np.zeros((max_read_length, 256), int)
    R2_q_array = np.zeros((max_read_length, MAX_EXPECTED_QUAL + 1), int)
    R2_c_array = np.zeros((max_read_length, 256), int)
    
    joint_average_q_distribution = np.zeros((MAX_EXPECTED_QUAL + 1, MAX_EXPECTED_QUAL + 1), int)
    
    for R1, R2 in read_pairs:
        R1_average_q = process_read(R1.seq.encode(), R1.qual.encode(), R1_q_array, R1_c_array)
        R2_average_q = process_read(R2.seq.encode(), R2.qual.encode(), R2_q_array, R2_c_array)
        joint_average_q_distribution[int(R1_average_q), int(R2_average_q)] += 1
        
    # See comment in quality_and_complexity above. 
    R1_c_array = np.vstack([R1_c_array.T[ord(b)] for b in base_order]).T
    R2_c_array = np.vstack([R2_c_array.T[ord(b)] for b in base_order]).T
    
    R1_average_q_distribution = joint_average_q_distribution.sum(axis=1) 
    R2_average_q_distribution = joint_average_q_distribution.sum(axis=0) 

    results = {
        'R1_qs': R1_q_array,
        'R1_cs': R1_c_array,
        'R2_qs': R2_q_array,
        'R2_cs': R2_c_array,
        'joint_average_qs': joint_average_q_distribution,
        'R1_average_qs': R1_average_q_distribution,
        'R2_average_qs': R2_average_q_distribution,
    }
    
    return results

def get_line_groups(line_source):
    if isinstance(line_source, Path):
        line_source = str(line_source)

    if isinstance(line_source, str):
        if line_source.endswith('.gz'):
            opener = functools.partial(gzip.open, mode='rt')
        else:
            opener = open
    else:
        opener = iter

    lines = opener(line_source)
    groups = zip(*[lines]*4)
    return groups

make_record = '@{0}\n{1}\n+\n{2}\n'.format

class Read(object):
    def __init__(self, name, seq, qual):
        self.name = name
        self.seq = seq
        self.qual = qual
        
    def __str__(self):
        return make_record(self.name, self.seq, self.qual)
    
    def reverse_complement(self):
        return Read(self.name,
                    utilities.reverse_complement(self.seq),
                    self.qual[::-1],
                   )
    
    def __getitem__(self, sl):
        return Read(self.name, self.seq[sl], self.qual[sl])
    
    def __repr__(self):
        return str(self)

    def __len__(self):
        return len(self.seq)

    def __add__(self, other):
        return Read(self.name, self.seq + other.seq, self.qual + other.qual)

    @property
    def query_sequence(self):
        return self.seq

    @property
    def query_name(self):
        return self.name

    @property
    def qname(self):
        return self.name
    
    @utilities.memoized_property
    def query_qualities(self):
        return array.array('B', decode_sanger(self.qual))

    @utilities.memoized_property
    def Q30_fraction(self):
        return np.mean(np.array(self.query_qualities) >= 30)

    @utilities.memoized_property
    def Q93_fraction(self):
        return np.mean(np.array(self.query_qualities) == 93)
    
def line_group_to_read(line_group, name_standardizer=identity, qual_convertor=identity):
    name_line, seq_line, _, qual_line = line_group

    name = name_standardizer(name_line.rstrip().lstrip('@'))

    seq = seq_line.strip()
    seq = seq.translate(period_to_N)

    qual = qual_convertor(qual_line.strip())

    return Read(name, seq, qual)

def get_read_length(file_name):
    read = next(reads(file_name))
    return len(read)

def reads(file_name, standardize_names=False, ensure_sanger_encoding=False, up_to_space=False):
    ''' Yields Read's from a file name or line iterator.
        If standardize_names == True, infers read name structure and
        standardizes read names.
        If ensure_sanger_encoding == True, detects the quality score encoding
        and converts to sanger if necessary.
    '''
    if isinstance(file_name, list):
        args = [standardize_names, ensure_sanger_encoding, up_to_space]
        files = [reads(fn, *args) for fn in file_name]
        return chain.from_iterable(files)

    line_groups = get_line_groups(file_name)

    if standardize_names:
        name_standardizer, line_groups = detect_structure(line_groups)
    elif up_to_space:
        name_standardizer = lambda n: n.split(' ')[0]
    else:
        name_standardizer = identity

    if ensure_sanger_encoding:
        qual_convertor, line_groups = detect_encoding(line_groups)
    else:
        qual_convertor = identity
    
    reads_ = (
        line_group_to_read(line_group, name_standardizer, qual_convertor)
        for line_group in line_groups
    )

    return reads_

def reverse_complement_reads(file_name, **kwargs):
    for read in reads(file_name, **kwargs):
        yield read.reverse_complement()

def detect_structure(line_groups):
    ''' Look at the first read to figure out the read name structure. '''
    try:
        first_group = next(line_groups)
        first_read = line_group_to_read(first_group)
        name_standardizer = get_read_name_standardizer(first_read.name)
        line_groups = chain([first_group], line_groups)
    except StopIteration:
        name_standardizer = identity
        # Note: line_groups is now an empty iterator.
    
    return name_standardizer, line_groups

def detect_encoding(line_groups):
    groups_examined = []
    try:
        for line_group in line_groups:
            groups_examined.append(line_group)
            read = line_group_to_read(line_group)
            ords = [ord(q) for q in read.qual]
            if min(ords) < SOLEXA_OFFSET - 5:
                encoding = 'SANGER'
                break
            if max(ords) > SANGER_OFFSET + 41:
                encoding = 'SOLEXA'
                break
    except StopIteration:
        encoding = 'SANGER'
        # Note: line_groups is now an empty iterator.

    if encoding == 'SOLEXA':
        qual_convertor = solexa_to_sanger
    else:
        qual_convertor = identity

    line_groups = chain(groups_examined, line_groups)
    
    return qual_convertor, line_groups

def read_pairs(R1_file_name, R2_file_name, **kwargs):
    R1_reads = reads(R1_file_name, **kwargs)
    R2_reads = reads(R2_file_name, **kwargs)
    return zip(R1_reads, R2_reads)

def read_pairs_interleaved(lines, **kwargs):
    interleaved_reads = reads(lines, **kwargs)
    grouped = group_by(interleaved_reads, key=lambda r: get_pair_name(r.name))
    for pair_name, group in grouped:
        if len(group) != 2:
            raise ValueError(group)
        R1, R2 = group
        R1_renamed = Read(pair_name, R1.seq, R1.qual)
        R2_renamed = Read(pair_name, R2.seq, R2.qual)
        yield R1_renamed, R2_renamed

quartet_order = ['I1', 'I2', 'R1', 'R2']
Quartet = namedtuple('Quartet', quartet_order)

def read_quartets(fns, **kwargs):
    if isinstance(fns, dict):
        fns = [fns[which] for which in quartet_order]

    all_reads = [reads(fn, **kwargs) for fn in fns]

    return (Quartet(*reads) for reads in zip(*all_reads))

def get_read_name_parser(read_name):
    if read_name.startswith('test'):
        # Simulated data sometimes needs read names to contain information
        # and can't use standard Illumina-formatted read names.
        parser = None
    elif read_name.startswith('SRR'):
        if len(read_name.split('.')) == 2:
            parser = parse_SRA_read_name
        elif len(read_name.split('.')) == 3:
            parser = parse_paired_SRA_read_name
    elif read_name.startswith('ERR') or read_name.startswith('DRR'):
        parser = parse_ERR_read_name
    else:
        num_words = len(read_name.split())
        if num_words > 1:
            parser = parse_new_illumina_read_name
        elif num_words == 1:
            if '#' in read_name:
                parser = parse_old_illumina_read_name
            elif '/' in read_name:
                parser = parse_unindexed_old_illumina_read_name
            else:
                parser = parse_standardized_name
        else:
            raise ValueError(f'read name format not recognized - {read_name}')
    return parser

def get_read_name_standardizer(read_name):
    ''' Looks at structure of read_name to determine the appropriate read name
        standardizer.
    '''
    parser = get_read_name_parser(read_name)
    if parser == parse_SRA_read_name or parser == parse_ERR_read_name:
        standardize = templates['SRA'].format
        def standardizer(read_name):
            accession, number = parser(read_name)
            standardized = standardize(accession, number)
            return standardized
    elif parser == parse_paired_SRA_read_name:
        standardize = templates['paired_SRA'].format
        def standardizer(read_name):
            accession, number, member = parser(read_name)
            standardized = standardize(accession, number, member)
            return standardized
    elif parser is not None:
        standardize = templates['default'].format
        def standardizer(read_name):
            lane, tile, x, y, member, index = parser(read_name)
            standardized = standardize(lane, tile, x, y)
            return standardized
    else:
        standardizer = identity

    return standardizer

def standardize(template, *args):
    return templates[template].format(*args)

templates = {
    'default': '{0:0>2.2s}:{1:0>5.5s}:{2:0>6.6s}:{3:0>6.6s}',
    'SRA': '{0:0>9.9s}:{1:0>10.10s}',
    'paired_SRA': '{0:0>9.9s}:{1:0>10.10s}:{2:0>1.1s}',
}

def parse_new_illumina_read_name(read_name):
    location_info, member_info = read_name.split()[:2]
    lane, tile, x, y = location_info.split(':')[-4:]
    member, _, _, index = member_info.split(':')
    return lane, tile, x, y, member, index

def parse_old_illumina_read_name(read_name):
    location_info, member_info = read_name.split('#')
    lane, tile, x, y = location_info.split(':')[-4:]
    index, member = member_info.split('/')
    return lane, tile, x, y, member, index

def parse_unindexed_old_illumina_read_name(read_name):
    location_info, member = read_name.split('/')
    lane, tile, x, y = location_info.split(':')[-4:]
    return lane, tile, x, y, member, ''

def parse_standardized_name(read_name):
    lane, tile, x, y, member = read_name.split(':')
    return lane, tile, x, y, member, ''

def parse_SRA_read_name(read_name):
    accession, number = read_name.split()[0].split('.')
    # Remove the leading 'SRR'
    accession = accession[3:]
    return accession, number

def parse_paired_SRA_read_name(read_name):
    accession, number, member = read_name.split('.')
    # Remove the leading 'SRR'
    accession = accession[3:]
    return accession, number, member

def parse_ERR_read_name(read_name):
    accession, number = read_name.split()[0].split('.')
    # Remove the leading 'SRR'
    accession = accession[3:]
    return accession, number

def coordinates_from_standardized(standardized):
    coordinates = standardized.split(':')[:-1]
    return coordinates

def get_pair_name(standardized):
    ''' Returns the part of a standardized name that is common to both members of
        the pair it came from.
    '''
    coordinates = coordinates_from_standardized(standardized)
    pair_name = ':'.join(coordinates)
    return pair_name



import logging
import copy
from collections import Counter

import numpy as np
import pysam

from . import sam
from . import utilities
from . import fastq
from . import interval

def keep_same_names(R1_aligned, R2_aligned):
    return R1_aligned.qname, R2_aligned.qname

def is_disoriented(R1_aligned, R2_aligned):
    disoriented = False

    if R1_aligned.is_reverse:
        if R2_aligned.is_reverse:
            disoriented = True
        if R1_aligned.pos < R2_aligned.pos:
            disoriented = True
    else:
        if not R2_aligned.is_reverse:
            disoriented = True
        if R2_aligned.pos < R1_aligned.pos:
            disoriented = True

    return disoriented

def reference_interval(R1_m, R2_m):
    if R1_m.tid != R2_m.tid:
        return None
    
    min_start = min(R1_m.reference_start, R2_m.reference_start)
    max_end = max(R1_m.reference_end, R2_m.reference_end)

    # reference_end points one past the end.
    return interval.Interval(min_start, max_end - 1)

def get_reference_extent(R1_m, R2_m):
    ref_interval = reference_interval(R1_m, R2_m)
    return len(ref_interval)

def is_discordant(R1_m, R2_m, max_insert_length):
    discordant = False
    if R1_m.tid != R2_m.tid:
        discordant = True
    else:
        tlen = get_reference_extent(R1_m, R2_m)
        if tlen > max_insert_length:
            discordant = True

    return discordant

def remove_soft_clipping(mapping):
    if mapping.cigar[0][0] == sam.BAM_CSOFT_CLIP:
        clipped_from_start = mapping.cigar[0][1]
        cigar_start_index = 1
        cigar_start = mapping.cigar[:1]
    else:
        clipped_from_start = 0
        cigar_start_index = 0
        cigar_start = []
    
    if mapping.cigar[-1][0] == sam.BAM_CSOFT_CLIP:
        clipped_from_end = mapping.cigar[-1][1]
        cigar_end_index = -1
        cigar_end = mapping.cigar[-1:]
    else:
        clipped_from_end = 0
        cigar_end_index = None
        cigar_end = []

    seq = mapping.seq
    qual = mapping.qual
    from_start_slice = slice(None, clipped_from_start)
    from_end_slice = slice(len(seq) - clipped_from_end, None)
    remaining_slice = slice(clipped_from_start, len(seq) - clipped_from_end)

    clipped = {
        'from_start': {
            'seq': seq[from_start_slice],
            'qual': qual[from_start_slice],
            'cigar': cigar_start,
        },
        'from_end': {
            'seq': seq[from_end_slice],
            'qual': qual[from_end_slice],
            'cigar': cigar_end,
        },
    }

    mapping.cigar = mapping.cigar[cigar_start_index:cigar_end_index]
    mapping.seq = seq[remaining_slice]
    mapping.qual = qual[remaining_slice]

    return clipped

def combine_paired_mappings(R1_mapping, R2_mapping, verbose=False):
    ''' Takes two pysam mappings representing opposite ends of a fragment and
    combines them into one mapping, (ab)using BAM_CREF_SKIP to bridge the gap
    (if any) between them.
    '''
    R1_mapping = copy.deepcopy(R1_mapping)
    R2_mapping = copy.deepcopy(R2_mapping)

    R1_strand = sam.get_strand(R1_mapping)

    if R1_strand == '+':
        left_mapping, right_mapping = R1_mapping, R2_mapping
    elif R1_strand == '-':
        left_mapping, right_mapping = R2_mapping, R1_mapping
                
    # Soft-clipping at the 3' end of a read should only happen if this is
    # read-through into soft-clipping at the 5' end of the other read.
    # If there is non-physical soft-clipping in this pair, give up now.
    # Specifically, check if any pairing of read position to ref position
    # isn't the same.

    if (left_mapping.cigar[-1][0] == sam.BAM_CSOFT_CLIP) or \
       (right_mapping.cigar[0][0] == sam.BAM_CSOFT_CLIP):

        left_pairs = left_mapping.get_aligned_pairs(matches_only=True)
        right_pairs = right_mapping.get_aligned_pairs(matches_only=True)

        if left_pairs != right_pairs:
            return False

    # Otherwise, remove all soft-clipping from the mappings, storing the 5'
    # soft-clipped seq and quals from both reads to add back at the end.
    
    left_clipped = remove_soft_clipping(left_mapping)
    right_clipped = remove_soft_clipping(right_mapping)

    left_md = dict(left_mapping.tags)['MD']
    right_md = dict(right_mapping.tags)['MD']

    right_aligned_pairs = sam.cigar_to_aligned_pairs(right_mapping.cigar, right_mapping.reference_start)

    right_after_overlap_pair_index = len(right_aligned_pairs)
    for i, (read, ref) in enumerate(right_aligned_pairs):
        if ref != None and ref >= left_mapping.aend:
            right_after_overlap_pair_index = i
            break
    
    right_overlap_pairs = right_aligned_pairs[:right_after_overlap_pair_index]
    right_after_overlap_pairs = right_aligned_pairs[right_after_overlap_pair_index:]
    
    right_reads_after = [read for read, ref in right_after_overlap_pairs if read != None and read != 's']
    right_refs_after = [ref for read, ref in right_after_overlap_pairs if ref != None]
    
    right_overlap_cigar = sam.aligned_pairs_to_cigar(right_overlap_pairs)
    right_after_overlap_cigar = sam.aligned_pairs_to_cigar(right_after_overlap_pairs)
    right_after_overlap_md = sam.truncate_md_string_from_beginning(right_md, len(right_refs_after))
    
    right_after_overlap_read_start = len(right_mapping.seq) - len(right_reads_after)

    right_overlap_seq = right_mapping.seq[:right_after_overlap_read_start] 
    right_overlap_qual = right_mapping.query_qualities[:right_after_overlap_read_start] 

    right_after_overlap_seq = right_mapping.seq[right_after_overlap_read_start:]
    right_after_overlap_qual = right_mapping.qual[right_after_overlap_read_start:]
    
    
    left_aligned_pairs = sam.cigar_to_aligned_pairs(left_mapping.cigar, left_mapping.reference_start)

    left_before_overlap_pair_index = -1
    for i, (read, ref) in list(enumerate(left_aligned_pairs))[::-1]:
        if ref != None and ref < right_mapping.pos:
            left_before_overlap_pair_index = i
            break

    left_overlap_pairs = left_aligned_pairs[left_before_overlap_pair_index + 1:]
    left_before_overlap_pairs = left_aligned_pairs[:left_before_overlap_pair_index + 1]

    left_reads_before = [read for read, ref in left_before_overlap_pairs if read != None and read != 's']
    left_refs_before = [ref for read, ref in left_before_overlap_pairs if ref != None]
    
    left_overlap_cigar = sam.aligned_pairs_to_cigar(left_overlap_pairs)
    left_before_overlap_cigar = sam.aligned_pairs_to_cigar(left_before_overlap_pairs)
    left_before_overlap_md = sam.truncate_md_string_up_to(left_md, len(left_refs_before))
    
    left_overlap_read_start = len(left_reads_before)
    left_overlap_seq = left_mapping.seq[left_overlap_read_start:] 
    left_overlap_qual = left_mapping.query_qualities[left_overlap_read_start:] 

    left_before_overlap_seq = left_mapping.seq[:left_overlap_read_start]
    left_before_overlap_qual = left_mapping.qual[:left_overlap_read_start]

    if left_overlap_pairs or right_overlap_pairs:
        gap_length = 0

        left_has_splicing = sam.contains_splicing(left_mapping)
        right_has_splicing = sam.contains_splicing(right_mapping)

        if left_overlap_cigar == right_overlap_cigar:
            # If the two mappings agree about the location of indels in their overlap,
            # use the seq from the mapping with the higher average quality in the
            # overlap.
            left_mean_qual = np.mean(left_overlap_qual)
            right_mean_qual = np.mean(right_overlap_qual)

            if left_mean_qual > right_mean_qual:
                use_overlap_from = 'left'
            else:
                use_overlap_from = 'right'
        elif left_has_splicing != right_has_splicing:
            # A temporary(?) heuristic - if one read has splicing and the other
            # doesn't, use the overlap from the one with splicing under the
            # assumption that the other just has a few bases overhanging the
            # splice junction.
            if left_has_splicing:
                use_overlap_from = 'left'
            else:
                use_overlap_from = 'right'
        else:
            # If the two mappings disagree about the location of indels in their overlap,
            # we need a heuristic for picking which mapping we believe reflects the
            # true structure of the input fragment. The most innocuous explanation
            # is that a 'true' indel happened to lie close to the edge of one of the
            # mappings. A more problematic situation is a 'false' indel (that is,
            # produced during cluster generation or sequencing-by-synthesis, NOT
            # template production). Our strategy is: realign the overlapping part of
            # left mapping starting from the left edge of the overlap according to the
            # cigar of the right mapping and realign the overlapping part of the right
            # mapping starting from the right edge of the overlap according to the cigar
            # of the left mapping. Count the number of mismatches produced by each.
            # If the left overlap can accomodate the right cigar with fewer mismatches,
            # use the right cigar and seq. If the right overlap can accomodate the left
            # cigar with fewer mismatches, use the left cigar and seq.

            # The leftmost aligned_pair from the right mapping is guaranteed by the
            # mapping process to not involve a gap.
            _, overlap_ref_start = right_overlap_pairs[0]
            # Similarly, the rightmost aligned_pair from the left mapping can't be a
            # gap.
            _, overlap_ref_end = left_overlap_pairs[-1]

            realigned_left_cigar = sam.truncate_cigar_blocks_up_to(right_mapping.cigar, len(left_overlap_seq))
            realigned_right_cigar = sam.truncate_cigar_blocks_from_beginning(left_mapping.cigar, len(right_overlap_seq))

            ref_dict = sam.merge_ref_dicts(sam.ref_dict_from_mapping(left_mapping),
                                           sam.ref_dict_from_mapping(right_mapping),
                                          )

            try:
                left_using_right_mismatches = realigned_mismatches(left_overlap_seq, overlap_ref_start, realigned_left_cigar, ref_dict)
                right_using_left_mismatches = realigned_mismatches_backwards(right_overlap_seq, overlap_ref_end, realigned_right_cigar, ref_dict)
            except (ValueError, TypeError):
                print(left_mapping)
                print(right_mapping)
                raise
            
            if verbose:
                logging.info('disagreements in {0}'.format(left_mapping.qname))
                logging.info('left overlap cigar is  {0}'.format(str(left_overlap_cigar)))
                logging.info('right overlap cigar is {0}'.format(str(right_overlap_cigar)))
                logging.info('left_using_right_mismatches - {0}'.format(len(left_using_right_mismatches)))
                logging.info('right_using_left_mismatches - {0}'.format(len(right_using_left_mismatches)))

            if len(left_using_right_mismatches) < len(right_using_left_mismatches):
                use_overlap_from = 'right'
            elif len(right_using_left_mismatches) < len(left_using_right_mismatches):
                use_overlap_from = 'left'
            else:
                logging.info('disagreements in {0}'.format(left_mapping.qname))
                logging.info('left overlap cigar is  {0}'.format(str(left_overlap_cigar)))
                logging.info('right overlap cigar is {0}'.format(str(right_overlap_cigar)))
                logging.info('left_using_right_mismatches - {0}'.format(len(left_using_right_mismatches)))
                logging.info('right_using_left_mismatches - {0}'.format(len(right_using_left_mismatches)))
                logging.info('ambiguous disagreement')
                return False

    else:
        gap_length = right_mapping.pos - left_mapping.aend
        # It doesn't matter what use_overlap_from is set to; there is no overlap
        use_overlap_from = 'left'
        
    combined_mapping = pysam.AlignedRead()
    combined_mapping.qname = left_mapping.qname
    combined_mapping.tid = left_mapping.tid
    combined_mapping.mapq = min(left_mapping.mapq, right_mapping.mapq)
    combined_mapping.rnext = -1
    combined_mapping.pnext = -1
    combined_mapping.pos = left_mapping.pos

    if R1_strand == '-':
        combined_mapping.is_reverse = True

    gap_cigar = [(sam.BAM_CREF_SKIP, gap_length)]
    
    if use_overlap_from == 'left':
        combined_mapping.seq = left_mapping.seq + right_after_overlap_seq
        combined_mapping.qual = left_mapping.qual + right_after_overlap_qual
        combined_mapping.cigar = left_mapping.cigar + gap_cigar + right_after_overlap_cigar
    
        combined_md = sam.combine_md_strings(left_md, right_after_overlap_md)
        combined_mapping.setTag('MD', combined_md)

        overlap_seq_tag = right_overlap_seq
        overlap_qual_tag = fastq.encode_sanger(right_overlap_qual)

    elif use_overlap_from == 'right':
        combined_mapping.seq = left_before_overlap_seq + right_mapping.seq
        combined_mapping.qual = left_before_overlap_qual + right_mapping.qual
        combined_mapping.cigar = left_before_overlap_cigar + gap_cigar + right_mapping.cigar

        combined_md = sam.combine_md_strings(left_before_overlap_md, right_md)
        combined_mapping.setTag('MD', combined_md)
        
        overlap_seq_tag = left_overlap_seq
        overlap_qual_tag = fastq.encode_sanger(left_overlap_qual)

    if len(overlap_seq_tag) > 0:
        # Having empty tags causes problems, so don't create them.
        combined_mapping.setTag('Xs', overlap_seq_tag)
        combined_mapping.setTag('Xq', overlap_qual_tag)
        combined_mapping.setTag('Xw', use_overlap_from)

    qual = combined_mapping.qual
    seq = combined_mapping.seq
    cigar = combined_mapping.cigar

    before = left_clipped['from_start']
    after = right_clipped['from_end']

    combined_mapping.cigar = before['cigar'] + cigar + after['cigar']
    combined_mapping.seq = before['seq'] + seq + after['seq']
    combined_mapping.qual = before['qual'] + qual + after['qual']

    return combined_mapping

def realigned_mismatches(seq, start, realigned_cigar, ref_dict):
    realigned_pairs = sam.cigar_to_aligned_pairs(realigned_cigar, start)
    mismatches = []
    for read_position, ref_position in realigned_pairs:
        if read_position != None and read_position != 's' and ref_position != None and ref_position != 'S':
            read_base = seq[read_position]
            ref_base = ref_dict[ref_position]
            if read_base != ref_base:
                mismatches.append((read_position, ref_position, read_base, ref_base))

    return mismatches

def realigned_mismatches_backwards(seq, end, realigned_cigar, ref_dict):
    realigned_pairs = sam.cigar_to_aligned_pairs_backwards(realigned_cigar, end, len(seq))
    mismatches = []
    for read_position, ref_position in realigned_pairs:
        if read_position != None and read_position != 's' and ref_position != None and ref_position != 'S':
            read_base = seq[read_position]
            ref_base = ref_dict[ref_position]
            if read_base != ref_base:
                mismatches.append((read_position, ref_position, read_base, ref_base))

    return mismatches

def find_skip_index_in_combined(combined_mapping):
    # There should be exactly one BAM_CREF_SKIP, possibly of length 0,
    # separating R1 from R2.
    skip_indices = [i for i, (op, length) in enumerate(combined_mapping.cigar)
                    if op == sam.BAM_CREF_SKIP]
    if len(skip_indices) != 1:
        raise ValueError(str(combined_mapping))
    skip_index = skip_indices.pop()
    return skip_index

def remove_length_zero_skip(combined_mapping):
    skip_index = find_skip_index_in_combined(combined_mapping)
    _, skip_length = combined_mapping.cigar[skip_index]
    if skip_length == 0:
        removed_cigar = remove_cigar_op(combined_mapping.cigar, skip_index)
    else:
        removed_cigar = combined_mapping.cigar
    return removed_cigar

def extract_seqs_from_combined(combined_mapping,
                               include_overlap=True,
                               remove_soft_clipped=True,
                               flip_if_reverse=True,
                              ):
    ''' Separates out the R1 and R2 seq and quals that went into a
    combined_mapping.
    '''
    strand = sam.get_strand(combined_mapping)
    tags = dict(combined_mapping.tags)
    if 'Xs' not in tags:
        tags['Xs'] = ''
        tags['Xq'] = ''
        tags['Xw'] = 'left'

    skip_index = find_skip_index_in_combined(combined_mapping)

    left_cigar = combined_mapping.cigar[:skip_index]
    right_cigar = combined_mapping.cigar[skip_index + 1:]

    left_length = sam.total_read_nucs(left_cigar)

    left_seq = combined_mapping.seq[:left_length]
    left_qual = combined_mapping.qual[:left_length]

    right_seq = combined_mapping.seq[left_length:]
    right_qual = combined_mapping.qual[left_length:]

    if remove_soft_clipped:
        first_left_op, first_left_length = left_cigar[0]
        if first_left_op == sam.BAM_CSOFT_CLIP:
            left_seq = left_seq[first_left_length:]
            left_qual = left_qual[first_left_length:]
        
        last_right_op, last_right_length = right_cigar[-1]
        if last_right_op == sam.BAM_CSOFT_CLIP:
            right_seq = right_seq[:-last_right_length]
            right_qual = right_qual[:-last_right_length]

    if include_overlap:
        if tags['Xw'] == 'left':
            # Overlapping sequence in the combined read reflects that from the 
            # left mapping, so the overlap from the right was stored in the Xs
            # and Xq tags.
            right_seq = tags['Xs'] + right_seq
            right_qual = tags['Xq'] + right_qual
        elif tags['Xw'] == 'right':
            # Overlapping sequence in the combined read reflects that from the 
            # right mapping, so the overlap from the left was stored in the Xs
            # and Xq tags.
            left_seq = left_seq + tags['Xs']
            left_qual = left_qual + tags['Xq']

    if strand == '+':
        R1_seq, R1_qual = left_seq, left_qual
        R2_seq, R2_qual = right_seq, right_qual

        if flip_if_reverse:
            R2_seq = utilities.reverse_complement(R2_seq)
            R2_qual = R2_qual[::-1]

    elif strand == '-':
        R1_seq, R1_qual = right_seq, right_qual
        R2_seq, R2_qual = left_seq, left_qual
        
        if flip_if_reverse:
            R1_seq = utilities.reverse_complement(R1_seq)
            R1_qual = R1_qual[::-1]

    return R1_seq, R1_qual, R2_seq, R2_qual

def extract_reads_from_combined(combined_mapping):
    R1_seq, R1_qual, R2_seq, R2_qual = extract_seqs_from_combined(combined_mapping, remove_soft_clipped=False)
    R1 = fastq.Read(combined_mapping.qname, R1_seq, R1_qual)
    R2 = fastq.Read(combined_mapping.qname, R2_seq, R2_qual)
    return R1, R2

def split_combined_mapping(combined_mapping, remove_soft_clipped=True):
    ''' Split a combined_mapping into non-overlapping mappings. '''
    R1_mapping = pysam.AlignedRead()
    R1_mapping.is_read1 = True
    R1_mapping.tid = combined_mapping.tid
    R1_mapping.qname = combined_mapping.qname

    R2_mapping = pysam.AlignedRead()
    R2_mapping.is_read2 = True
    R2_mapping.tid = combined_mapping.tid
    R2_mapping.qname = combined_mapping.qname

    skip_index = find_skip_index_in_combined(combined_mapping)

    left_cigar = combined_mapping.cigar[:skip_index]
    right_cigar = combined_mapping.cigar[skip_index + 1:]

    if remove_soft_clipped:
        first_left_op, first_left_length = left_cigar[0]
        if first_left_op == sam.BAM_CSOFT_CLIP:
            left_cigar = left_cigar[1:]
        
        last_right_op, last_right_length = right_cigar[-1]
        if last_right_op == sam.BAM_CSOFT_CLIP:
            right_cigar = right_cigar[:-1]
    
    combined_md = dict(combined_mapping.tags)['MD']
    left_ref_bases = sam.total_reference_nucs(left_cigar)
    right_ref_bases = sam.total_reference_nucs(right_cigar)

    _, gap = combined_mapping.cigar[skip_index]
    left_pos = combined_mapping.pos
    right_pos = left_pos + left_ref_bases + gap

    left_md = sam.truncate_md_string_up_to(combined_md, left_ref_bases)
    right_md = sam.truncate_md_string_from_beginning(combined_md, right_ref_bases)
    
    strand = sam.get_strand(combined_mapping)
    if strand == '+':
        R1_mapping.cigar = left_cigar
        R1_mapping.setTag('MD', left_md)
        
        R2_mapping.cigar = right_cigar
        R2_mapping.setTag('MD', right_md)

        R1_mapping.pos = left_pos
        R2_mapping.pos = right_pos

        R2_mapping.is_reverse = True
    elif strand == '-':
        R1_mapping.cigar = right_cigar
        R1_mapping.setTag('MD', right_md)
        
        R2_mapping.cigar = left_cigar
        R2_mapping.setTag('MD', left_md)

        R1_mapping.pos = right_pos
        R2_mapping.pos = left_pos
        R1_mapping.is_reverse = True

    R1_seq, R1_qual, R2_seq, R2_qual = extract_seqs_from_combined(combined_mapping,
                                                                  include_overlap=False,
                                                                  remove_soft_clipped=remove_soft_clipped,
                                                                  flip_if_reverse=False,
                                                                 )
    if R1_seq != '':
        R1_mapping.seq = R1_seq
        R1_mapping.qual = R1_qual
    if R2_seq != '':
        R2_mapping.seq = R2_seq
        R2_mapping.qual = R2_qual
    
    return R1_mapping, R2_mapping

def remove_cigar_op(cigar, index):
    op, length = cigar[index]
    if length != 0:
        raise ValueError('Attempted to remove cigar op with nonzero length')

    if index == 0:
        removed = cigar[1:]
    elif index == len(cigar) - 1:
        removed = ciger[:-1]
    else:
        before_op, before_length = cigar[index - 1]
        after_op, after_length = cigar[index + 1]
        if before_op == after_op:
            interface = [(before_op, before_length + after_length)]
        else:
            interface = [(before_op, before_length), (after_op, after_length)]
        removed = cigar[:index - 1] + interface + cigar[index + 2:]

    return removed

def mapping_from_line(line):
    combined_mapping = pysam.AlignedRead()
    parsed_line = sam.parse_line(line)
    if parsed_line['strand'] == '-':
        combined_mapping.is_reverse = True
    combined_mapping.seq = parsed_line['SEQ']
    combined_mapping.qual = parsed_line['QUAL']
    combined_mapping.cigarstring = parsed_line['CIGAR']
    
    # This should obviously be made more general.
    for tag_name in ['Xs', 'Xq', 'Xw']:
        if tag_name in parsed_line:
            tag = [(tag_name, parsed_line[tag_name])]
            combined_mapping.tags = combined_mapping.tags + tag 

    return combined_mapping

def filter_mappings(mappings,
                    minimum_mapq=42,
                    max_insert_length=1000,
                    counts_dict=None,
                    verbose=False,
                    unmapped_fns=None,
                   ):
    ''' Filters out unmapped, nonuniquely mapped, or discordantly mapped
        reads.
    '''
    pair_counts = {'total': 0,
                   'unmapped': 0,
                   'indel': 0,
                   'nonunique': 0,
                   'discordant': 0,
                   'disoriented': 0,
                   'unique': Counter(),
                   'mapqs': Counter(),
                   'fragment_lengths': Counter(),
                   'tids': Counter(),
                  }

    if unmapped_fns:
        R1_unmapped_fn, R2_unmapped_fn = unmapped_fns
        R1_unmapped_fh = open(R1_unmapped_fn, 'w')
        R2_unmapped_fh = open(R2_unmapped_fn, 'w')

    for _, aligned_pair in utilities.group_by(mappings, key=lambda m: m.qname):
        if len(aligned_pair) != 2:
            raise ValueError(len(aligned_pair))

        pair_counts['total'] += 1
        
        R1_aligned, R2_aligned = aligned_pair
        # If R2 is mapped but R1 isn't, R2 gets reported first.
        if not R1_aligned.is_read1:
            R1_aligned, R2_aligned = R2_aligned, R1_aligned

        if (not R1_aligned.is_read1) or (not R2_aligned.is_read2):
            raise ValueError(R1_aligned, R2_aligned)
        
        pair_counts['mapqs'][R1_aligned.mapq] += 1
        pair_counts['mapqs'][R2_aligned.mapq] += 1

        if R1_aligned.is_unmapped or R2_aligned.is_unmapped:
            pair_counts['unmapped'] += 1
            
            if verbose:
                logging.info('{0} was unmapped'.format(R1_aligned.qname))
            
            if unmapped_fns:
                R1_read = sam.mapping_to_Read(R1_aligned)
                R2_read = sam.mapping_to_Read(R2_aligned)
                R1_unmapped_fh.write(str(R1_read))
                R2_unmapped_fh.write(str(R2_read))
        
        elif is_discordant(R1_aligned, R2_aligned, max_insert_length):
            pair_counts['discordant'] += 1
        
        else:
            pair_counts['tids'][R1_aligned.tid] += 1

            if is_disoriented(R1_aligned, R2_aligned):
                pair_counts['disoriented'] += 1
            elif R1_aligned.mapq < minimum_mapq or R2_aligned.mapq < minimum_mapq:
                pair_counts['nonunique'] += 1
                if verbose:
                    logging.info('{0} was nonunique, {1}, {2}'.format(R1_aligned.qname, R1_aligned.mapq, R2_aligned.mapq))
            else:
                pair_counts['unique'][R1_aligned.tid] += 1

                fragment_length = abs(R1_aligned.tlen)
                pair_counts['fragment_lengths'][fragment_length] += 1
                
                if sam.contains_indel(R1_aligned) or sam.contains_indel(R2_aligned):
                    pair_counts['indel'] += 1
                
                yield R1_aligned, R2_aligned

    if counts_dict != None:
        counts_dict.update(pair_counts)

def extract_fragment_lengths(bam_file_name):
    def concordantly_mapped(mapping):
        return not mapping.mate_is_unmapped and \
               not mapping.is_unmapped and \
               mapping.rnext == mapping.tid and \
               abs(mapping.tlen) < 10000

    bam_fh = pysam.Samfile(bam_file_name)
    TLENs = (abs(mapping.tlen) for mapping in bam_fh if mapping.is_read1 and concordantly_mapped(mapping))
    fragment_lengths = Counter(TLENs)
    
    # Note that counts_to_array implicitly discards negative key values.
    fragment_lengths = utilities.counts_to_array(fragment_lengths)

    return fragment_lengths

def filter_long_TLENs(bam_file_name, filtered_bam_file_name, max_TLEN):
    bam_file = pysam.Samfile(bam_file_name)
    filtered_bam_file = pysam.Samfile(filtered_bam_file_name, 'wb', template=bam_file)
    for aligned_read in bam_file:
        if aligned_read.tlen < max_TLEN:
            filtered_bam_file.write(aligned_read)

def get_concordant_pairs(R1_group, R2_group, max_insert_length):
    ''' Results are sorted by reference extent length. '''
    pairs = []
    for R1_m in R1_group:
        for R2_m in R2_group:
            if not is_discordant(R1_m, R2_m, max_insert_length) and not is_disoriented(R1_m, R2_m):
                pairs.append((R1_m, R2_m))

    pairs = sorted(pairs, key=lambda p: get_reference_extent(*p))

    return pairs

def group_mapping_pairs(mappings):
    groups = utilities.group_by(mappings, lambda r: r.query_name)
    for query_name, query_mappings in groups:
        R1_group = [m for m in query_mappings if m.is_read1]
        R2_group = [m for m in query_mappings if m.is_read2]
        yield query_name, (R1_group, R2_group)



from collections import Counter

def _read_gene(fh, name_line):
    name, num_lines = name_line.strip().split()
    num_lines = int(num_lines)
    gene = Counter()
    for i in range(num_lines):
        key, count = fh.readline().strip().split()
        count = int(count)
        x, y = map(int, key.split(','))
        gene[x, y] = count

    return name, gene

def _write_gene(name, gene, fh):
    name_line = '{0}\t{1}\n'.format(name, len(gene))
    fh.write(name_line)
    for (x, y), count in gene.most_common():
        fh.write('{0},{1}\t{2}\n'.format(x, y, count))

def read_file(file_name):
    genes = {}
    fh = open(file_name)
    name_line = fh.readline()
    while name_line:
        name, gene = _read_gene(fh, name_line)
        genes[name] = gene
        name_line = fh.readline()

    return genes

def write_file(genes, file_name):
    with open(file_name, 'w') as fh:
        for name in sorted(genes):
            _write_gene(name, genes[name], fh)

def combine_data(first_genes, second_genes):
    for name in second_genes:
        if name not in first_genes:
            first_genes[name] = second_genes[name]
        else:
            first_genes[name].update(second_genes[name])

    return first_genes



import subprocess
import os
import contextlib
import sys

import yaml
import hits.parallel

def piece(srr_fn, num_pieces, which_piece, paired=False):
    root, _ = os.path.splitext(srr_fn)
    yaml_fn = '{0}.yaml'.format(root)

    if not os.path.exists(yaml_fn):
        raise ValueError('yaml doesn\'t exist for {0}'.format(srr_fn))

    with open(yaml_fn) as yaml_fh:
        info = yaml.load(yaml_fh)

    total_spots = info['total_spots']

    bounds = hits.parallel.get_bounds(total_spots, num_pieces)
    first = bounds[which_piece] + 1
    last = bounds[which_piece + 1]

    with dump_spots(srr_fn, first, last, paired) as lines:
        for line in lines:
            # Note: empirically, these lines are bytes objects.
            yield line.decode()

@contextlib.contextmanager
def dump_spots(srr_fn, first, last, paired):
    if paired:
        name_format = '@$ac.$si.$ri'
    else:
        name_format = '@$ac.$si'

    command = [
        'fastq-dump',
        '--dumpbase',
        '--minSpotId', str(first),
        '--maxSpotId', str(last),
        '--defline-seq', name_format,
        '--stdout',
        srr_fn,
    ]

    if paired:
        command.insert(1, '--split-spot')

    process = subprocess.Popen(command,
                               stdout=subprocess.PIPE,
                               stderr=subprocess.PIPE,
                              )
    lines = iter(process.stdout)
    try:
        yield lines
    finally:
        process.terminate()
        process.stdout.close()
        for line in process.stderr:
            if not (line.startswith(b'Read') or line.startswith(b'Written')):
                sys.stderr.write(line)



import glob
import os
from collections import namedtuple

import pysam
import Bio.SeqIO

def get_all_fasta_file_names(genome_dir):
    fasta_file_names = [
        fn for fn in glob.glob('{0}/*.fa*'.format(genome_dir))
        if not fn.endswith('.fai') and not 'fastq' in fn
    ]
    return fasta_file_names

def get_all_fai_file_names(genome_dir):
    fai_file_names = [fn for fn in glob.glob('{0}/*.fai'.format(genome_dir))]
    return fai_file_names

def make_fais(genome_dir):
    fasta_file_names = get_all_fasta_file_names(genome_dir)
    for fn in fasta_file_names:
        pysam.faidx(fn)

fai_entry_fields = [
    'file_name',
    'length',
    'offset',
    'bases_per_line',
    'bytes_per_line',
]
fai_entry = namedtuple('fai_entry', fai_entry_fields)

def parse_fai(fai_file_name):
    fasta_file_name, _ = os.path.splitext(fai_file_name) 

    def parse_line(line):
        fields = line.strip().split()
        seq_name = fields[0]
        values = map(int, fields[1:])
        return seq_name, fai_entry(fasta_file_name, *values)

    entries = {}

    with open(fai_file_name) as fh:
        for line in fh:
            k, v = parse_line(line)
            entries[k] = v

    return entries

def get_genome_index(genome_dir):
    fai_file_names = get_all_fai_file_names(genome_dir)
    entries = {}
    for fai_file_name in fai_file_names:
        entries.update(parse_fai(fai_file_name))
    return entries

def build_base_lookup(genome_dir):
    ''' Returns a memoized function for looking up single bases from reference.
    '''
    genome_index = get_genome_index(genome_dir)
    references = {}

    def base_lookup(rname, position):
        if position < 0:
            raise IndexError

        if rname not in references:
            fasta_file_name = genome_index[rname].file_name
            with pysam.Fastafile(fasta_file_name) as fasta_file:
                references[rname] = fasta_file.fetch(reference=rname)

        base = references[rname][position:position + 1]
        return base

    return base_lookup

def build_region_fetcher(genome_dir, load_references=False):
    ''' Returns a function for fetching regions from the genome in genome_dir.
        If load_references == True, loads entire reference sequences into memory
        the first time they are fetched from.
        If the returned function is given a negative start or an end that is
        longer than the seq_name's sequence, the region returned will be
        padded with -.
    '''
    genome_index = get_genome_index(genome_dir)

    fasta_files = {
        fasta_file_name: pysam.Fastafile(fasta_file_name)
        for fasta_file_name in get_all_fasta_file_names(genome_dir)
    }

    seq_name_to_file = {
        seq_name: fasta_files[genome_index[seq_name].file_name]
        for seq_name in genome_index
    }

    references = {}
    def lookup_loaded(seq_name, start, end):
        if seq_name not in references:
            references[seq_name] = seq_name_to_file[seq_name].fetch(seq_name)

        region = references[seq_name][start:end]
        return region

    def lookup_unloaded(seq_name, start, end):
        region = seq_name_to_file[seq_name].fetch(seq_name, start, end)
        return region

    if load_references:
        lookup = lookup_loaded
    else:
        lookup = lookup_unloaded

    def region_fetcher(seq_name, start, end):
        if end < 0:
            return '-'*(end - start)
        if start < 0:
            left_pad = '-'*(-start)
            start = 0
        else:
            left_pad = ''

        region = lookup(seq_name, start, end)
        right_pad = '-'*((end - start) - len(region))

        return left_pad + region + right_pad

    return region_fetcher

def load_entire_genome(genome_dir):
    seqs = {}
    fasta_file_names = get_all_fasta_file_names(genome_dir)
    for fasta_file_name in fasta_file_names:
        for record in Bio.SeqIO.parse(fasta_file_name, 'fasta'):
            seqs[record.id] = str(record.seq)
    return seqs

def max_RNAME_length(genome_index):
    length = max(len(name) for name in genome_index)
    return length

def max_POS_length(genome_index):
    length = max(len(str(genome_index[name].length)) for name in genome_index)
    return length

def max_tid_length(genome_index):
    return len(str(len(genome_index)))



from pathlib import Path

import bokeh.models

def load_js_template(file, key):
    fn = Path(file).parent / f'{key}.js'
    return fn.read_text()

def build_js_callback(file, key, format_kwargs=None, args=None):
    if args is None:
        args = {}
    if format_kwargs is None:
        format_kwargs = {}

    code = load_js_template(file, key)
    for format_kwarg, value in format_kwargs.items():
        to_replace = '{' + format_kwarg + '}'
        code = code.replace(to_replace, str(value))

    name = f'{key}_callback'
    callback = bokeh.models.CustomJS(code=code, args=args)
    callback.name = name

    return callback


import struct

MAGIC_STRING = b'BUS\x00'
PAD = 0

def read_header(fh):
    magic_string = fh.read(4)
    if magic_string != MAGIC_STRING:
        raise ValueError(magic_string)
        
    version, = struct.unpack('I', fh.read(4))
    barcode_length, = struct.unpack('I', fh.read(4))
    umi_length, = struct.unpack('I', fh.read(4))
    
    header_length, = struct.unpack('I', fh.read(4))
    header = fh.read(header_length)

    return barcode_length, umi_length

def write_header(fh, barcode_length, umi_length):
    version = 1
    fh.write(MAGIC_STRING)
    fh.write(struct.pack('I', version))
    fh.write(struct.pack('I', barcode_length))
    fh.write(struct.pack('I', umi_length))

    header = b'written by hits'
    fh.write(struct.pack('I', len(header)))
    fh.write(header)

i_to_b = 'ACGT'
b_to_i = {b: i for i, b in enumerate(i_to_b)}

def twobit_to_seq(n, length):
    seq = ''.join(i_to_b[(n >> (2 * (length - 1 - i))) & 3] for i in range(length))
    return seq

def seq_to_twobit(seq):
    b = seq[0]
    n = b_to_i[b]
    
    for b in seq[1:]:
        n <<= 2
        i = b_to_i[b]
        n |= i
        
    return n

RecordStruct = struct.Struct('LLiIII')

class Record():
    def __init__(self, barcode, umi, eq_class, count, flags):
        self.barcode = barcode
        self.umi = umi
        self.eq_class = eq_class
        self.count = count
        self.flags = flags

    @classmethod
    def from_file_handle(cls, fh, barcode_length, umi_length):
        bs = fh.read(RecordStruct.size)
        if len(bs) != RecordStruct.size:
            if len(bs) > 0:
                raise ValueError('malformed file')
            else:
                raise EOFError

        barcode, umi, eq_class, count, flags, pad = RecordStruct.unpack(bs)
        return cls(twobit_to_seq(barcode, barcode_length),
                   twobit_to_seq(umi, umi_length),
                   eq_class,
                   count,
                   flags,
                  )

    def pack(self):
        bs = RecordStruct.pack(seq_to_twobit(self.barcode),
                               seq_to_twobit(self.umi),
                               self.eq_class,
                               self.count,
                               self.flags,
                               PAD,
                              )
        return bs

def records(fn):
    with open(fn, 'rb') as fh:
        barcode_length, umi_length = read_header(fh)
        try:
            while True:
                record = Record.from_file_handle(fh, barcode_length, umi_length)
                yield record
        except EOFError:
            pass

def write_file(fn, records, barcode_length, umi_length):
    with open(fn, 'wb') as fh:
        write_header(fh, barcode_length, umi_length)
        for record in records:
            fh.write(record.pack())


import functools
import operator
from collections import defaultdict
from numbers import Number

import numpy as np

from . import sam

def are_disjoint(first, second):
    if first.is_empty or second.is_empty:
        return True
    else:
        return first.start > second.end or second.start > first.end

def are_overlapping(first, second):
    return not are_disjoint(first, second)

def are_adjacent(first, second):
    return first.start == second.end + 1 or second.start == first.end + 1

class Interval:
    def __init__(self, start, end):
        self.start = start
        self.end = end
        self.is_empty = (end < start)

    @classmethod
    def from_feature(self, feature):
        ''' only really necessary to set is_empty '''
        return Interval(feature.start, feature.end)

    @classmethod
    def empty(self):
        ''' Note: evalutates via bool to False since length is 0. '''
        return Interval(-1, -2)
    
    @classmethod
    def from_slice(self, sl):
        if sl.start == None:
            start = 0
        else:
            start = sl.start

        if sl.stop == None:
            end = np.inf
        else:
            end = sl.stop - 1 # Note the -1

        return Interval(start, end)

    @property
    def slice(self):
        return slice(self.start, self.end + 1)

    def __or__(self, other):
        if are_disjoint(self, other):
            left, right = sorted([self, other])
            if are_adjacent(self, other):
                intervals = [Interval(left.start, right.end)]
            else:
                intervals = [left, right]
        else:
            intervals = [Interval(min(self.start, other.start), max(self.end, other.end))]
            
        return DisjointIntervals(intervals)
    
    def __and__(self, other):
        if isinstance(other, DisjointIntervals):
            # Defer to definition in DisjointIntervals
            return other & self
        elif are_disjoint(self, other):
            return Interval.empty()
        else:
            return Interval(max(self.start, other.start), min(self.end, other.end))
        
    def __contains__(self, other):
        if isinstance(other, Interval):
            # is a strict sub-interval of
            return (other.start >= self.start and other.end <= self.end) and (self != other)
        elif isinstance(other, Number):
            return self.start <= other <= self.end
        else:
            raise ValueError(other)
        
    @property    
    def comparison_key(self):
        return self.start, self.end

    @property
    def total_length(self):
        return len(self)
    
    def __lt__(self, other):
        return self.comparison_key < other.comparison_key
    
    def __repr__(self):
        return '[{0:,} - {1:,}]'.format(self.start, self.end)
    
    def __key(self):
        return (self.start, self.end)

    def __eq__(self, other):
        return self.__key() == other.__key()

    def __hash__(self):
        return hash(self.__key())
    
    def __ne__(self, other):
        return not self == other

    def __len__(self):
        if self.is_empty:
            return 0
        else:
            return self.end - self.start + 1

    def __sub__(self, other):
        if isinstance(other, DisjointIntervals):
            survived_all = DisjointIntervals([self])
            for other_interval in other:
                survived_this = self - other_interval
                survived_all = survived_all & survived_this
            return survived_all

        else:
            left = Interval(self.start, min(self.end, other.start - 1))
            right = Interval(max(self.start, other.end + 1), self.end)
            disjoint = DisjointIntervals([left, right])

            if disjoint.total_length == 0:
                return Interval(-1, -2)
            elif len(disjoint.intervals) == 1:
                return disjoint.intervals[0]
            else:
                return disjoint

    def __iter__(self):
        return iter(range(self.start, self.end + 1))
    
class DisjointIntervals:
    def __init__(self, intervals):
        ''' Note that nothing enforces disjointness here - if
        that is what is needed, use make_disjoint(intervals)
        '''
        self.intervals = sorted([i for i in intervals if i.end >= i.start])
        self.is_empty = (len(self.intervals) == 0)
        
    def __len__(self):
        return len(self.intervals)
    
    @property
    def start(self):
        if len(self.intervals) == 0:
            return None
        else:
            return min(interval.start for interval in self.intervals)
    
    @property
    def end(self):
        if len(self.intervals) == 0:
            return None
        else:
            return max(interval.end for interval in self.intervals)

    @property
    def span(self):
        ''' The convext hull of the intervals. '''
        if len(self.intervals) == 0:
            return Interval.empty()
        else:
            return Interval(self.start, self.end)
        
    def __repr__(self):
        return '{{{}}}'.format(', '.join(map(str, self.intervals)))
    
    def __getitem__(self, sl):
        return self.intervals[sl]
    
    def __or__(self, other):
        if isinstance(other, DisjointIntervals):
            everything = self
            for other_interval in other:
                everything = everything | other_interval
            return everything
        else:
            disjoints = []
            
            for interval in self.intervals:
                union = interval | other
                if len(union) > 1:
                    disjoints.append(interval)
                else:
                    other = union[0]
                    
            disjoints.append(other)
            
            return DisjointIntervals(sorted(disjoints))
    
    def __and__(self, other):
        if isinstance(other, DisjointIntervals):
            survived_some = DisjointIntervals([])
            for other_interval in other:
                survived_this = self & other_interval
                survived_some = survived_some | survived_this
            return survived_some
        else:
            intersections = []
            for interval in self.intervals:
                intersection = interval & other
                if intersection:
                    intersections.append(intersection)
                
            return DisjointIntervals(intersections)
    
    def __eq__(self, other):
        if type(other) == Interval:
            return len(self.intervals) == 1 and self.intervals[0] == other
        else:
            return self.intervals == other.intervals

    def __iter__(self):
        return iter(self.intervals)

    def __hash__(self):
        return hash(self.intervals)
    
    def __ne__(self, other):
        return not self == other
    
    def __contains__(self, other):
        if isinstance(other, Number):
            other = Interval(other, other)

        return (self | other) == self

    @property
    def total_length(self):
        return sum(len(i) for i in self.intervals)
    
    def __sub__(self, other):
        if isinstance(other, DisjointIntervals):
            return functools.reduce(operator.and_, [self - other_int for other_int in other.intervals], self)
        else:
            pieces = []
            for interval in self.intervals:
                piece = interval - other
                if isinstance(piece, Interval):
                    pieces.append(piece)
                elif isinstance(piece, DisjointIntervals):
                    pieces.extend(piece)

            return DisjointIntervals(pieces)

def get_covered(alignment):
    if alignment is None or alignment.is_unmapped:
        return Interval(-1, -2)
    else:
        return Interval(*sam.query_interval(alignment))

def get_covered_on_ref(alignment):
    if alignment is None or alignment.is_unmapped:
        return Interval(-1, -2)
    else:
        return Interval(*sam.ref_interval(alignment))

def make_disjoint(intervals):
    disjoint = DisjointIntervals([])
    for interval in intervals:
        disjoint = disjoint | interval
    return disjoint

def get_disjoint_covered(alignments):
    intervals = [get_covered(al) for al in alignments if al is not None]
    covered = make_disjoint(intervals)
    return covered

def remove_nested(alignments):
    unnecessary = set()
    covered_list = [get_covered(al) for al in alignments]
    for i, left in enumerate(covered_list):
        for j, right in enumerate(covered_list):
            if i == j:
                continue
            if left in right:
                unnecessary.add(i)
    necessary = [al for i, al in enumerate(alignments) if i not in unnecessary]
    return necessary

def make_parsimonious(alignments):
    initial_covered = get_disjoint_covered(alignments)
    
    no_nested = remove_nested(alignments)
    interval_to_als = defaultdict(list)
    for al in no_nested:
        interval_to_als[get_covered(al)].append(al)
        
    unique_intervals = sorted(interval_to_als, key=len, reverse=True)
    remaining = unique_intervals
        
    contributes = []
    for possibly_exclude in unique_intervals:
        exclude_one = [intvl for intvl in remaining if intvl != possibly_exclude]
        now_covered = make_disjoint(exclude_one)
        if initial_covered != now_covered:
            contributes.append(possibly_exclude)
        else:
            remaining = exclude_one
            
    parsimonious = []
    for interval in contributes:
        parsimonious.extend(interval_to_als[interval])
    
    if get_disjoint_covered(parsimonious) != initial_covered:
        raise ValueError
        
    return parsimonious



import itertools
import numpy as np

def get_bounds(length, num_pieces):
    base_size = length // num_pieces
    num_with_extra = length - base_size * num_pieces
    sizes = np.ones(num_pieces, int) * base_size
    sizes[:num_with_extra] += 1
    bounds = np.append([0], sizes.cumsum())
    return bounds

def piece_of_list(full_list, num_pieces, which_piece, interleaved=False):
    if which_piece == -1:
        # Sentinel value that means 'the whole thing'.
        piece = full_list
    elif interleaved:
        piece = [x for i, x in enumerate(full_list) if i % num_pieces == which_piece]
    else:
        bounds = get_bounds(len(full_list), num_pieces)
        piece = full_list[bounds[which_piece]:bounds[which_piece + 1]]

    return piece

def piece_of_iter(full_iter, num_pieces, which_piece):
    if which_piece == -1:
        # Sentinel value that means 'the whole thing'.
        piece = full_iter
    else:
        piece = itertools.islice(full_iter, which_piece, None, num_pieces)

    return piece



import numbers
import base64
import re
from collections import defaultdict
from pathlib import Path

from matplotlib.pyplot import hist

import numpy as np
import scipy.cluster.hierarchy
import bokeh
import bokeh.io
import bokeh.plotting
import pandas as pd
import matplotlib.colors
import matplotlib.cm
import IPython.display

bokeh.io.output_notebook()

def bool_to_js(b):
    return 'true' if b else 'false'

from ..callback import build_js_callback

def scatter(df=None,
            numerical_cols=None,
            hover_keys=None,
            table_keys=None,
            color_by=None,
            label_by=None,
            size=800,
            label_size=8,
            axis_label_size=20,
            log_scale=False,
            volcano=False,
            heatmap=False,
            cluster=False,
            grid='none',
            marker_size=6,
            initial_selection=None,
            initial_xy_names=None,
            initial_alpha=0.5,
            nonselection_alpha=0.1,
            initial_data_lims=None,
            initial_x_lims=None,
            initial_y_lims=None,
            data_bounds=None,
            zoom_to_initial_data=False,
            alpha_widget_type='slider',
            hide_widgets=None,
            identical_bins=False,
            num_bins=100,
            show_axes_lines=False,
            return_layout=False,
            title=None,
            two_level_index=False,
            menu_width=200,
            save_as=None,
            level_order=None,
           ):
    ''' Makes an interactive scatter plot using bokeh. Call without any
    arguments for an example using data from Jan et al. Science 2014.

    Args:
        df: A pandas DataFrame with columns containing numerical data to plot. 
            Index values will be used as labels for points.
            Any text columns will be searchable through the 'Search:' field.
            Any boolean columns will be used to define subsets of points for
            selection from a dropdown menu.
            If df is None, loads example data from Jan et al. Science 2014.

        numerical_cols: If given, a list of columns to use as plotting choices.
            (If not given, all columns containing numerical data will be used.)

        hover_keys: Names of columns in df to display in the tooltip that
            appears when you hover over a point.

        table_keys: Names of columns in df to display in the table below the
            plot that is populated with the selected points from the figure.

        color_by: The name of a column in df to use as colors of points, or a
            list of such names to choose from a menu. (These columns will be
            excluded from text searches.)
        
        label_by: The name of a column in df to use as labels of points, or a
            list of such names to choose from a menu. If None, df.index is used.

        size: Size of the plot in pixels.

        marker_size: Size of the scatter circles. Can be a scalar value, a
            column name, or a list of column names to choose from via dropdown
            menu.

        label_size: Size in pts of font for labels of selected point.

        heatmap: If True, displays a heatmap of correlations between numerical
            columns in df that can be clicked to select columns to scatter. If
            any negative correlations exist, uses a red/blue colormap, otherwise
            uses viridis.

        cluster: If True, forces heatmap=True and performs hierarchical
            clustering on (correlations of) numerical columns and draws a
            dendrogram.

        grid: Draw a 'grid', 'diagonal' lines, 'diagonal+axes', or 'none' as guide lines.

        volcano: If True, make some tweaks suitable for volcano plots.

        log_scale: If not False, plot on a log scale with base 10 (or, if a set
            to a number, with base log_scale.)

        axis_label_size: Size of the font used for axis labels.

        initial_selection: Names of index value to initially highlight.

        initial_xy_names: Tuple (x_name, y_name) of datasets to initially
            display on x- and y-axes.

        alpha_widget_type: Type of widget ('slider' or 'text') to control
            scatter circle transparency.

        hide_widgets: List of widgets to not display. Possible options are
            ['table', 'alpha', 'marker_size', 'search', 'subset_menu',
             'grid_radio_buttons'].

        identical_bins: If True, use the same set of bins for histograms of all
            data sets. If False, use bins specific to the range of each data
            set.

        num_bins: Number of bins to use for marginal histograms.

        zoom_to_initial_data: If True, zoom to data limits of the initially
            selected columns rather than global data limits.

        return_layout: If True, return the final layout object to allowing
            embedding.
    '''

    if save_as is not None:
        bokeh.io.output_file(str(save_as))

    if hover_keys is None:
        hover_keys = []

    if table_keys is None:
        table_keys = []

    if level_order is None:
        level_order = {}

    if hide_widgets is None:
        hide_widgets = set()
    else:
        hide_widgets = set(hide_widgets)

    if volcano:
        grid = 'grid'

    if cluster:
        heatmap = True

    if df is None:
        # Load example data.
        fn = Path(__file__).parent / 'example_df.txt'
        df = pd.read_csv(fn, index_col='alias')

        # Override some arguments.
        log_scale = True
        hover_keys = ['systematic_name', 'short_description']
        table_keys = ['systematic_name', 'description']
        grid = 'diagonal'
        heatmap = False
        identical_bins = True

    # Copy before changing.
    original_index_name = df.index.name
    original_df = df
    df = df.copy()

    # Collapse multiindex if present
    df.columns = [' '.join(map(str, n)) if isinstance(n, tuple) else n for n in df.columns]
    df.index = [' '.join(map(str, n)) if isinstance(n, tuple) else n for n in df.index.values]
    df.index.name = original_index_name

    # Infer column types.
    scatter_data = df.to_dict(orient='list')

    if 'index' in scatter_data:
        scatter_data['_index'] = scatter_data['index']

    if df.index.name is None:
        df.index.name = 'index'
    
    scatter_data[df.index.name] = list(df.index)

    if initial_selection is None:
        initial_selection = []

    initial_indices = [i for i, n in enumerate(df.index) if n in initial_selection]

    if two_level_index:
        two_level_numerical_cols = list(original_df.select_dtypes('number').columns)

    numerical_cols = list(df.select_dtypes('number').columns)

    # bokeh can handle NaNs in numpy arrays but not in lists. 
    for numerical_col in numerical_cols:
        scatter_data[numerical_col] = np.array(scatter_data[numerical_col])

    object_cols = [n for n in df.columns if df[n].dtype is np.dtype('O')]
    if df.index.dtype is np.dtype('O'):
        object_cols.append(df.index.name)

    bool_cols = [n for n in df.columns if df[n].dtype is np.dtype('bool')]

    subset_indices = {n: [i for i, v in enumerate(df[n]) if v] for n in bool_cols}

    # Set up the actual scatter plot.
    
    tools = [
        'reset',
        'undo',
        'pan',
        'box_zoom',
        'box_select',
        'tap',
        'wheel_zoom',
        'save',
    ]

    if isinstance(size, tuple):
        plot_width, plot_height = size
    else:
        plot_width, plot_height = size, size

    if two_level_index:
        plot_width += 35 # empirical adjustment for extra blank text row

    fig_kwargs = dict(
        plot_width=plot_width,
        plot_height=plot_height,
        tools=tools,
        lod_threshold=80000,
        name='scatter_fig',
        toolbar_location=None,
    )

    min_border = 80

    if log_scale:
        if log_scale is True:
            log_scale = 10
        fig_kwargs['y_axis_type'] = 'log'
        fig_kwargs['x_axis_type'] = 'log'
    
    fig = bokeh.plotting.figure(**fig_kwargs)
    fig.toolbar.logo = None

    if log_scale:
        for axis in [fig.xaxis, fig.yaxis]:
            axis[0].ticker.base = log_scale
            axis[0].formatter.ticker = axis[0].ticker

    fig.grid.visible = (grid == 'grid')
    fig.grid.name = 'grid'
    
    lasso = bokeh.models.LassoSelectTool(select_every_mousemove=False)
    fig.add_tools(lasso)
    
    if initial_xy_names is None:
        if two_level_index:
            first_level = original_df.columns.levels[0]
            second_level = [s for s in original_df.columns.levels[1] if s != '']

            if first_level[1] != 'gene':
                initial_xy_names = (
                    (first_level[0], second_level[0]),
                    (first_level[1], second_level[0]),
                )
            else:
                initial_xy_names = (
                    (first_level[0], second_level[0]),
                    (first_level[0], second_level[1]),
                )
        else:
            initial_xy_names = numerical_cols[:2]

    if two_level_index:
        x_name, y_name = [' '.join(t) for t in initial_xy_names]
    else:
        x_name, y_name = initial_xy_names

    fig.xaxis.name = 'x_axis'
    fig.yaxis.name = 'y_axis'

    if two_level_index:
        x_title = bokeh.models.Title(text=initial_xy_names[0][0], text_font_size='16pt', align='center')
        x_subtitle = bokeh.models.Title(text=initial_xy_names[0][1], text_font_size='14pt', align='center')
        fig.add_layout(x_title, 'below')
        fig.add_layout(x_subtitle, 'below')

        y_title = bokeh.models.Title(text=initial_xy_names[1][0], text_font_size='16pt', align='center')
        y_subtitle = bokeh.models.Title(text=initial_xy_names[1][1], text_font_size='14pt', align='center')

        fig.add_layout(bokeh.models.Title(text=' ', text_font_size='6pt'), 'left')
        fig.add_layout(y_subtitle, 'left')
        fig.add_layout(y_title, 'left')
    else:
        fig.xaxis.axis_label = x_name
        fig.yaxis.axis_label = y_name
        for axis in (fig.xaxis, fig.yaxis):
            axis.axis_label_text_font_size = f'{axis_label_size}pt'
            axis.axis_label_text_font_style = 'normal'

    scatter_data['x'] = scatter_data[x_name]
    scatter_data['y'] = scatter_data[y_name]
    
    scatter_data['index'] = list(df.index)

    scatter_data['_black'] = ['rgba(0, 0, 0, 1.0)' for _ in scatter_data['x']]
    scatter_data['_orange'] = ['orange' for _ in scatter_data['x']]
    
    if color_by is None:
        color_by = ''
        show_color_by_menu = False
        color_options = ['']

        scatter_data['_color'] = scatter_data['_black']
        scatter_data['_selection_color'] = scatter_data['_orange']
    
    else:
        show_color_by_menu = 'color_by' not in hide_widgets

        if isinstance(color_by, str):
            color_options = ['', color_by]
        else:
            color_options = [''] + list(color_by)

        scatter_data['_color'] = scatter_data[color_options[-1]]
        scatter_data['_selection_color'] = scatter_data[color_options[-1]]
    
    if label_by is None:
        label_by = df.index.name

    if isinstance(label_by, str):
        show_label_by_menu = False
        label_options = [label_by]
    else:
        show_label_by_menu = True
        label_options = list(label_by)
    
    scatter_data['_label'] = scatter_data[label_options[0]]

    if isinstance(marker_size, numbers.Number):
        size_widget_type = 'slider'
    else:
        if isinstance(marker_size, str):
            size_options = ['', marker_size]
        else:
            size_options = [''] + list(marker_size)
    
        size_widget_type = 'menu'
        marker_size = '_size'
        scatter_data[marker_size] = scatter_data[size_options[1]]

        scatter_data['_uniform_size'] = [6]*len(scatter_data[marker_size])

    scatter_source = bokeh.models.ColumnDataSource(data=scatter_data,
                                                   name='scatter_source',
                                                  )
    if len(initial_indices) > 0:
        scatter_source.selected.indices = initial_indices

    scatter = fig.scatter('x',
                          'y',
                          source=scatter_source,
                          size=marker_size,
                          fill_color='_color',
                          fill_alpha=initial_alpha,
                          line_color=None,
                          selection_color='_selection_color',
                          selection_alpha=0.9,
                          nonselection_color='_color',
                          nonselection_alpha=nonselection_alpha,
                          name='scatter',
                         )
    
    if log_scale:
        nonzero = df[numerical_cols][df[numerical_cols] > 0]

        overall_max = nonzero.max().max()
        overall_min = nonzero.min().min()
        
        initial = (overall_min * 0.1, overall_max * 10)
        bounds = (overall_min * 0.001, overall_max * 1000)
    
        def log(x):
            return np.log(x) / np.log(log_scale)

        bins = {}
        for name in numerical_cols:
            if identical_bins:
                left = overall_min * 0.9
                right = overall_max / 0.9
            else:
                name_min = nonzero[name].min()
                name_max = nonzero[name].max()
                left = name_min * 0.9
                right = name_max / 0.9

            bins[name] = list(np.logspace(log(left), log(right), num_bins, base=log_scale))

    else:
        overall_max = df.max(numeric_only=True).max()
        overall_min = df.min(numeric_only=True).min()
        overall_buffer = (overall_max - overall_min) * 0.05
        
        extent = overall_max - overall_min
        overhang = extent * 0.05
        max_overhang = extent * 0.5

        initial = (overall_min - overhang, overall_max + overhang)
        bounds = (overall_min - max_overhang, overall_max + max_overhang)
        
        bins = {}
        for name in numerical_cols:
            if identical_bins:
                left = overall_min - overall_buffer
                right = overall_max + overall_buffer
            else:
                name_min = df[name].min()
                name_max = df[name].max()
                name_buffer = (name_max - name_min) * 0.05
                left = name_min - name_buffer
                right = name_max + name_buffer
            
            bins[name] = list(np.linspace(left, right, num_bins))

    if initial_x_lims is None:
        if initial_data_lims is None:
            initial_x_lims = initial
        else:
            initial_x_lims = initial_data_lims

    if initial_y_lims is None:
        if initial_data_lims is None:
            initial_y_lims = initial
        else:
            initial_y_lims = initial_data_lims

    if data_bounds is not None:
        bounds = data_bounds
        # Ensure that auto-data limits don't extend out of bounds.
        initial_x_lims = (max(initial_x_lims[0], bounds[0]), min(initial_x_lims[1], bounds[1]))
        initial_y_lims = (max(initial_y_lims[0], bounds[0]), min(initial_y_lims[1], bounds[1]))

    diagonals_visible = ('diagonal' in grid)

    if log_scale:
        upper_ys = np.array(bounds) * 10
        lower_ys = np.array(bounds) * 0.1
    else:
        upper_ys = np.array(bounds) + 1
        lower_ys = np.array(bounds) - 1

    line_kwargs = dict(
        color='black',
        nonselection_color='black',
        alpha=0.4,
        nonselection_alpha=0.4,
    ) 

    diagonal = fig.line(x=bounds, y=bounds, **line_kwargs)
    lines = [
        diagonal,
    ]

    for line in lines:
        line.visible = diagonals_visible
    
    axes_lines = [
        fig.line(x=bounds, y=0, name='axes_line', **line_kwargs),
        fig.line(x=0, y=bounds, name='axes_line', **line_kwargs),
    ]

    for line in axes_lines:
        line.visible = show_axes_lines or ('axes' in grid)
    
    if volcano:
        fig.y_range = bokeh.models.Range1d(-0.1, 8)
        fig.x_range = bokeh.models.Range1d(-1, 1)
    else:
        if zoom_to_initial_data:
            x_min, x_max = bins[x_name][0], bins[x_name][-1]
            y_min, y_max = bins[y_name][0], bins[y_name][-1]
        else:
            x_min, x_max = initial_x_lims
            y_min, y_max = initial_y_lims

        fig.y_range = bokeh.models.Range1d(y_min, y_max)
        fig.x_range = bokeh.models.Range1d(x_min, x_max)

    fig.x_range.name = 'x_range'
    fig.y_range.name = 'y_range'
    
    lower_bound, upper_bound = bounds
    range_kwargs = dict(lower_bound=lower_bound, upper_bound=upper_bound)
    
    fig.outline_line_color = 'black'

    scatter.selection_glyph.line_color = None
    scatter.nonselection_glyph.line_color = None

    # Make marginal histograms.

    histogram_data = {}
    histogram_data = {
        'zero': [0]*(num_bins - 1),
    }

    for name in numerical_cols:
        histogram_data.update({
            f'{name}_bins_left': bins[name][:-1],
            f'{name}_bins_right': bins[name][1:],
        })

    max_count = 0
    for name in numerical_cols:
        counts, _ = np.histogram(df[name].dropna(), bins=bins[name])
        max_count = max(max(counts), max_count)
        histogram_data[f'{name}_all'] = list(counts)

    if log_scale:
        axis_type = 'log'
    else:
        axis_type = 'linear'

    hist_figs = {
        'x': bokeh.plotting.figure(width=plot_width, height=100,
                                   x_range=fig.x_range,
                                   x_axis_type=axis_type,
                                   name='hists_x',
                                  ),
        'y': bokeh.plotting.figure(width=100, height=plot_height,
                                   y_range=fig.y_range,
                                   y_axis_type=axis_type,
                                   name='hists_y',
                                   min_border_bottom=60 + (41 if two_level_index else 0), # mysterious behavior
                                  ),
    }

    if title is None:
        title = ''

    hist_figs['x'].title.text = title

    for axis, name in [('x', x_name), ('y', y_name)]:
        for data_type in ['all', 'bins_left', 'bins_right']:
            axis_key = f'{axis}_{data_type}'
            name_key = f'{name}_{data_type}'
            histogram_data[axis_key] = histogram_data[name_key]

        initial_vals = df[name].iloc[initial_indices]
        initial_counts, _ = np.histogram(initial_vals.dropna(), bins[name])
    
        histogram_data[f'{axis}_selected'] = initial_counts

    histogram_source = bokeh.models.ColumnDataSource(data=histogram_data,
                                                     name='histogram_source',
                                                    )
    initial_hist_alpha = 0.1 if len(initial_indices) > 0 else 0.2
    quads = {}
    quads['x_all'] = hist_figs['x'].quad(left='x_bins_left',
                                         right='x_bins_right',
                                         bottom='zero',
                                         top='x_all',
                                         source=histogram_source,
                                         color='black',
                                         alpha=initial_hist_alpha,
                                         line_color=None,
                                         name='hist_x_all',
                                        )

    quads['x_selected'] = hist_figs['x'].quad(left='x_bins_left',
                                              right='x_bins_right',
                                              bottom='zero',
                                              top='x_selected',
                                              source=histogram_source,
                                              color='orange',
                                              alpha=0.8,
                                              line_color=None,
                                             )

    quads['y_all'] = hist_figs['y'].quad(top='y_bins_left',
                                         bottom='y_bins_right',
                                         left='zero',
                                         right='y_all',
                                         source=histogram_source,
                                         color='black',
                                         alpha=initial_hist_alpha,
                                         line_color=None,
                                         name='hist_y_all',
                                        )

    quads['y_selected'] = hist_figs['y'].quad(top='y_bins_left',
                                              bottom='y_bins_right',
                                              left='zero',
                                              right='y_selected',
                                              source=histogram_source,
                                              color='orange',
                                              alpha=0.8,
                                              line_color=None,
                                             )

    # Poorly-understood bokeh behavior causes selection changes to sometimes
    # be broadcast to histogram_source. To prevent this from having any visual
    # effect, remove any difference in selection/nonselection glyphs.
    for quad in quads.values():
        quad.selection_glyph = quad.glyph
        quad.nonselection_glyph = quad.glyph

    if identical_bins:
        x_end = max_count
        y_end = max_count
    else:
        x_end = max(histogram_source.data['x_all'])
        y_end = max(histogram_source.data['y_all'])

    hist_figs['x'].y_range = bokeh.models.Range1d(name='hist_x_range',
                                                  start=0,
                                                  end=x_end,
                                                  bounds='auto',
                                                 )
    hist_figs['y'].x_range = bokeh.models.Range1d(name='hist_y_range',
                                                  start=0,
                                                  end=y_end,
                                                  bounds='auto',
                                                 )

    for hist_fig in hist_figs.values():
        hist_fig.outline_line_color = None
        hist_fig.axis.visible = False
        hist_fig.grid.visible = False
        hist_fig.min_border = 0
        hist_fig.toolbar_location = None

    # Configure tooltips that pop up when hovering over a point.
    
    hover = bokeh.models.HoverTool(renderers=[scatter])
    hover.tooltips = [
        (df.index.name, '@{{{0}}}'.format(df.index.name)),
    ]
    for key in hover_keys:
        hover.tooltips.append((key, '@{{{0}}}'.format(key)))
    fig.add_tools(hover)

    # Set up the table.

    table_col_names = [df.index.name] + table_keys
    columns = []
    for col_name in table_col_names:
        lengths = [len(str(v)) for v in scatter_source.data[col_name]]
        mean_length = np.mean(lengths)

        if col_name in numerical_cols:
            formatter = bokeh.models.widgets.NumberFormatter(format='0.00')
            width = 50
        elif 'gene' in col_name:
            formatter = bokeh.models.widgets.HTMLTemplateFormatter(template='<a href="https://www.genecards.org/cgi-bin/carddisp.pl?gene=<%= value %>" target="_blank"><%= value %></a>')
            width = 100
        else:
            formatter = bokeh.models.StringFormatter()
            width = min(500, int(12 * mean_length))

        column = bokeh.models.widgets.TableColumn(field=col_name,
                                                  title=col_name,
                                                  formatter=formatter,
                                                  width=width,
                                                 )
        columns.append(column)

    filtered_data = {k: [scatter_source.data[k][i] for i in initial_indices]
                     for k in scatter_source.data
                    }
    
    filtered_source = bokeh.models.ColumnDataSource(data=filtered_data, name='filtered_source')
    
    table = bokeh.models.widgets.DataTable(source=filtered_source,
                                           columns=columns,
                                           width=2 * plot_width if heatmap else plot_width,
                                           height=600,
                                           sortable=False,
                                           reorderable=False,
                                           name='table',
                                           index_position=None,
                                          )
    
    labels = bokeh.models.LabelSet(x='x',
                                   y='y',
                                   text='_label',
                                   level='glyph',
                                   x_offset=0,
                                   y_offset=2,
                                   source=filtered_source,
                                   text_font_size=f'{label_size}pt',
                                   name='labels',
                                  )
    fig.add_layout(labels)
    
    # Set up menus or heatmap to select columns from df to put on x- and y-axis.

    if heatmap:
        correlations = df[numerical_cols].corr()
        correlations_array = np.array(correlations)
        if correlations_array.min() > 0:
            v_min = 0
            v_max = 1
            v_min = correlations_array.min()
            v_max = correlations_array.max()
            c_map = matplotlib.cm.viridis
        else:
            v_min = -1
            v_max = 1
            c_map = matplotlib.cm.RdBu_r

        norm = matplotlib.colors.Normalize(vmin=v_min, vmax=v_max)
        def r_to_color(r):
            color = matplotlib.colors.rgb2hex(c_map(norm(r)))
            return color
        
        linkage = scipy.cluster.hierarchy.linkage(correlations)
        dendro = scipy.cluster.hierarchy.dendrogram(linkage,
                                                    no_plot=True,
                                                    labels=numerical_cols,
                                                   )
        orders = {
            'original': numerical_cols,
            'clustered': dendro['ivl'],
        }
        data = defaultdict(list)

        for y, row in enumerate(numerical_cols):
            for x, col in enumerate(numerical_cols):
                for order_key, order in orders.items():
                    row = order[y]
                    col = order[x]

                    r = correlations[row][col]
                    data[f'r_{order_key}'].append(r)
                    data[f'x_{order_key}'].append(x)
                    data[f'x_name_{order_key}'].append(col)
                    data[f'y_{order_key}'].append(y)
                    data[f'y_name_{order_key}'].append(row)
                    data[f'color_{order_key}'].append(r_to_color(r))

        if cluster:
            order_key = 'clustered'
        else:
            order_key = 'original'

        for k in ['r', 'x', 'x_name', 'y', 'y_name', 'color']:
            data[k] = data[f'{k}_{order_key}']

        heatmap_source = bokeh.models.ColumnDataSource(data, name='heatmap_source')
        num_exps = len(numerical_cols)
        heatmap_size = size - 100
        heatmap_fig = bokeh.plotting.figure(tools='tap',
                                            x_range=(-0.5, num_exps - 0.5),
                                            y_range=(num_exps - 0.5, -0.5),
                                            width=heatmap_size, height=heatmap_size,
                                            toolbar_location=None,
                                            name='heatmap_fig',
                                           )

        heatmap_fig.grid.visible = False
        rects = heatmap_fig.rect(x='x', y='y',
                                 line_color=None,
                                 hover_line_color='black',
                                 hover_fill_color='color',
                                 selection_fill_color='color',
                                 nonselection_fill_color='color',
                                 nonselection_fill_alpha=1,
                                 nonselection_line_color=None,
                                 selection_line_color='black',
                                 line_width=5,
                                 fill_color='color',
                                 source=heatmap_source,
                                 width=1, height=1,
                                )
        for axis in [heatmap_fig.xaxis, heatmap_fig.yaxis]:
            axis.name = 'heatmap_axis'

        hover = bokeh.models.HoverTool()
        hover.tooltips = [
            ('X', '@x_name'),
            ('Y', '@y_name'),
            ('r', '@r'),
        ]
        heatmap_fig.add_tools(hover)

        def make_tick_formatter(order):
            return '''var dict = {dict}; return dict[tick].slice(0, 15);'''.format(dict=dict(enumerate(order)))
        
        for ax in [heatmap_fig.xaxis, heatmap_fig.yaxis]:
            ax.ticker = bokeh.models.FixedTicker(ticks=np.arange(num_exps))
            ax.formatter = bokeh.models.FuncTickFormatter(code=make_tick_formatter(orders[order_key]))
            ax.major_label_text_font_size = '8pt'

        heatmap_fig.xaxis.major_label_orientation = np.pi / 4

        # Turn off black lines on bottom and left.
        for axis in (heatmap_fig.xaxis, heatmap_fig.yaxis):
            axis.axis_line_color = None

        name_pairs = list(zip(heatmap_source.data['x_name'], heatmap_source.data['y_name']))
        initial_index = name_pairs.index((x_name, y_name))
        heatmap_source.selected.indices = [initial_index]

        heatmap_fig.min_border = 1
        
        dendro_fig = bokeh.plotting.figure(height=100, width=heatmap_size,
                                           x_range=heatmap_fig.x_range,
                                          )
        icoord = np.array(dendro['icoord'])
        # Want to covert the arbitrary scaling that icoord is given in so that
        # the min value maps to 0 and the max value maps to len(correlations) - 1
        interval = (icoord.max() - icoord.min()) / (len(correlations) - 1)
        xs = list((icoord - icoord.min()) / interval)
        dcoord = np.array(dendro['dcoord'])
        ys = list(dcoord)
        dendrogram_lines = dendro_fig.multi_line(xs, ys, color='black')

        if not cluster:
            dendrogram_lines.visible = False

        dendro_fig.outline_line_color = None
        dendro_fig.axis.visible = False
        dendro_fig.grid.visible = False
        dendro_fig.min_border = 0
        dendro_fig.toolbar_location = None
        dendro_fig.y_range = bokeh.models.Range1d(start=0, end=dcoord.max() * 1.05)
    
        # Button to toggle clustering.
        cluster_button = bokeh.models.widgets.Toggle(label='cluster heatmap',
                                                     width=50,
                                                     name='cluster_button',
                                                     active=cluster,
                                                    )

    elif two_level_index:
        if not isinstance(menu_width, tuple):
            menu_width = (menu_width, menu_width)

        menus = {}

        for level_index, width in zip((0, 1), menu_width):
            level_name = original_df.columns.names[level_index]
            if level_name is None:
                level_name = level_index

            if level_name in level_order:
                options = level_order[level_name]
            else:
                options = []
                for value_tuple in two_level_numerical_cols:
                    value = value_tuple[level_index]
                    if value not in options:
                        options.append(value)

            for names, axis in zip(initial_xy_names, ('x', 'y')):
                menus[axis, level_index] = bokeh.models.widgets.MultiSelect(title=f'{axis.upper()}-axis: {level_name}',
                                                                      options=options,
                                                                      value=[names[level_index]],
                                                                      size=min(10, len(options)),
                                                                      name=f'{axis}_{level_index}_menu',
                                                                      width=width,
                                                                     )

    else:
        x_menu = bokeh.models.widgets.MultiSelect(title='X',
                                                  options=numerical_cols,
                                                  value=[x_name],
                                                  size=min(6, len(numerical_cols)),
                                                  name='x_menu',
                                                 )
        y_menu = bokeh.models.widgets.MultiSelect(title='Y',
                                                  options=numerical_cols,
                                                  value=[y_name],
                                                  size=min(6, len(numerical_cols)),
                                                  name='y_menu',
                                                 )
        

    if not heatmap:
        # Just a placeholder so that we can assume cluster_button exists.
        cluster_button = bokeh.models.widgets.Toggle(name='cluster_button')

    # Button to toggle labels.
    label_button = bokeh.models.widgets.Toggle(label='Label selected points',
                                               width=50,
                                               active=True,
                                               name='label_button',
                                              )
    
    # Button to zoom to current data limits.
    zoom_to_data_button = bokeh.models.widgets.Button(label='Zoom to data limits',
                                                      width=50,
                                                      name='zoom_button',
                                                     )
    # Menu to choose label source.
    label_menu = bokeh.models.widgets.Select(title='Label by:',
                                             options=label_options,
                                             value=label_options[0],
                                             name='label_menu',
                                            )

    # Menu to choose color source.
    color_menu = bokeh.models.widgets.MultiSelect(title='Color by:',
                                                  options=color_options,
                                                  value=[color_options[-1]],
                                                  name='color_menu',
                                                 )

    # Radio group to choose whether to draw a vertical/horizontal grid or
    # diagonal guide lines. 
    options = ['grid', 'diagonal', 'diagonal+axes', 'none']
    active = options.index(grid)
    grid_options = bokeh.models.widgets.RadioGroup(labels=options,
                                                   active=active,
                                                   name='grid_radio_buttons',
                                                  )

    search_input = bokeh.models.widgets.TextInput(title='Search:')

    case_sensitive = bokeh.models.widgets.CheckboxGroup(labels=['Case sensitive'],
                                                        active=[],
                                                        name='case_sensitive',
                                                       )

    # Menu to select a subset of points from a columns of bools.
    subset_options = [''] + bool_cols
    subset_menu = bokeh.models.widgets.Select(title='Select subset:',
                                              options=subset_options,
                                              value='',
                                              name='subset_menu',
                                             )

    # button to dump table to file.
    save_button = bokeh.models.widgets.Button(label='Save table to file',
                                              width=50,
                                              name='save_button',
                                             )
    
    # button to dump annotations to file.
    annotation_button = bokeh.models.widgets.Button(label='save annotations to file',
                                                    width=50,
                                                    name='annotation_button',
                                                   )
    
    if alpha_widget_type == 'slider':
        alpha_widget = bokeh.models.Slider(start=0.,
                                           end=1.,
                                           value=initial_alpha,
                                           step=.05,
                                           title='alpha',
                                           name='alpha',
                                          )
    elif alpha_widget_type == 'text':
        alpha_widget = bokeh.models.TextInput(title='alpha', name='alpha', value=str(initial_alpha))
    else:
        raise ValueError('{0} not a valid alpha_widget_type value'.format(alpha_widget_type))

    if size_widget_type == 'slider':
        size_widget = bokeh.models.Slider(start=1,
                                          end=20.,
                                          value=marker_size,
                                          step=1,
                                          title='marker size',
                                          name='marker_size',
                                         )
    else:
        size_widget = bokeh.models.widgets.Select(title='Size by:',
                                                  options=size_options,
                                                  value=size_options[1],
                                                  name='marker_size',
                                                 )
    
    rect_source = bokeh.models.ColumnDataSource(data={
        'x': [],
        'y': [],
        'width': [],
        'height': [],
    }, name='rect_source')

    rects = fig.rect('x', 'y', 'width', 'height',
                     source=rect_source,
                     fill_alpha=0,
                     line_color='black',
                     line_alpha=0.5,
                    )
    rect_annotater = bokeh.models.BoxEditTool(renderers=[rects])
    if 'annotation' not in hide_widgets:
        fig.add_tools(rect_annotater)
    
    cross_source = bokeh.models.ColumnDataSource(data={
        'x': [],
        'y': [],
    }, name='cross_source')

    crosses = fig.cross('x', 'y', size=25,
                        line_color='black',
                        source=cross_source,
                       )
    cross_annotater = bokeh.models.PointDrawTool(renderers=[crosses])
    if 'annotation' not in hide_widgets:
        fig.add_tools(cross_annotater)

    fig.min_border = 1

    widgets = [
        label_button,
        zoom_to_data_button,
        cluster_button,
        label_menu,
        color_menu,
        alpha_widget,
        size_widget,
        grid_options,
        search_input,
        case_sensitive,
        subset_menu,
        save_button,
        annotation_button,
    ]

    if 'table' in hide_widgets:
        hide_widgets.add('save_button')

    if 'search' in hide_widgets:
        hide_widgets.add('case_sensitive')

    if 'annotation' in hide_widgets:
        hide_widgets.add('annotation_button')

    if not heatmap:
        hide_widgets.add('cluster_button')

    if not show_color_by_menu:
        hide_widgets.add('color_menu')

    if not show_label_by_menu:
        hide_widgets.add('label_menu')
    
    if len(subset_options) == 1:
        hide_widgets.add('subset_menu')

    widgets = [w for w in widgets if w.name not in hide_widgets]

    toolbar = bokeh.models.ToolbarBox(toolbar=fig.toolbar, toolbar_location='right')

    row = bokeh.layouts.row
    col = bokeh.layouts.column
    spacer = bokeh.layouts.Spacer

    columns = [
        col(hist_figs['x'], fig),
        col(spacer(height=100), hist_figs['y']),
        # Note: update to bokeh 2.4.1 changed behavior of ToolbarBox height
        # such that without this phantom 1px spacer, it thinks it can
        # only be 100px high and collapses most buttons into an overflow.
        col(spacer(height=100), row(spacer(width=1, height=800), toolbar)),
    ]

    if heatmap:
        heatmap_column = col(dendro_fig, heatmap_fig)
        columns.append(heatmap_column)
        widget_box = col(*widgets)
        final_column = col(spacer(height=min_border),
                           widget_box,
                          )
    else:
        if not two_level_index:
            widgets = [x_menu, y_menu] + widgets
            widget_box = col(*widgets)
            final_column = col(spacer(height=min_border),
                               widget_box,
                              )
        else:
            x_row = row(menus['x', 0], menus['x', 1])
            y_row = row(menus['y', 0], menus['y', 1])
            widget_box = col(*widgets)
            final_column = col(spacer(height=min_border),
                               x_row, y_row,
                               widget_box,
                              )

    columns.append(final_column)
        
    rows = [
        row(*columns),
    ]
    if 'table' not in hide_widgets:
        rows.append(table)

    full_layout = col(*rows)

    #rows = [
    #    hist_figs['x'],
    #    row(fig, hist_figs['y'], toolbar),
    #]

    #full_layout = col(*rows)


    # Build callbacks last so that all args are available.

    # Prevent ranges from going outside bounds.
    range_callback = build_js_callback(__file__, 'scatter_range',
                                       args=dict(
                                           x_range=fig.x_range,
                                           y_range=fig.y_range,
                                       ),
                                       format_kwargs=range_kwargs,
                                      )
    for range_ in [fig.x_range, fig.y_range]:
        for prop in ['start', 'end']:
            range_.js_on_change(prop, range_callback)

    # Filter the table when selection changes.
    selection_callback = build_js_callback(__file__, 'scatter_selection',
                                           args=dict(
                                               search_input=search_input,
                                               hist_x_all=quads['x_all'],
                                               hist_y_all=quads['y_all'],
                                               scatter_source=scatter_source,
                                               filtered_source=filtered_source,
                                               histogram_source=histogram_source,
                                               table=table,
                                               subset_menu=subset_menu
                                           ),
                                          )
    scatter_source.selected.js_on_change('indices', selection_callback)

    # Update displayed data when menu or heatmap selection changes.
    # Note: uses selection_callback, so must be after it.
    if heatmap:
        heatmap_callback = build_js_callback(__file__, 'scatter_heatmap',
                                             args=dict(
                                                 x_axis=fig.xaxis,
                                                 y_axis=fig.yaxis,
                                                 scatter_source=scatter_source,
                                                 filtered_source=filtered_source,
                                                 histogram_source=histogram_source,
                                                 heatmap_source=heatmap_source,
                                             ),
                                            )
        heatmap_source.selected.js_on_change('indices', heatmap_callback)

        # Cluster button
        format_kwargs = {
            # The callback expects code surrounded by double quotes.
            'clustered_formatter': '"{}"'.format(make_tick_formatter(orders['clustered'])),
            'original_formatter': '"{}"'.format(make_tick_formatter(orders['original'])),
        }
        cluster_callback = build_js_callback(__file__, 'scatter_cluster_button',
                                            format_kwargs=format_kwargs,
                                            args=dict(
                                                dendrogram_lines=dendrogram_lines,
                                                heatmap_source=heatmap_source,
                                                heatmap_x_axis=heatmap_fig.xaxis,
                                                heatmap_y_axis=heatmap_fig.yaxis,
                                                x_axis=fig.xaxis,
                                                y_axis=fig.yaxis,
                                                fig=fig,
                                            ),
                                            )
        cluster_button.js_on_click(cluster_callback)

    elif two_level_index:
        menu_callback = build_js_callback(__file__, 'scatter_two_level_menu',
                                          args=dict(
                                              scatter_source=scatter_source,
                                              filtered_source=filtered_source,
                                              histogram_source=histogram_source,
                                              x_menu_0=menus['x', 0],
                                              x_menu_1=menus['x', 1],
                                              y_menu_0=menus['y', 0],
                                              y_menu_1=menus['y', 1],
                                              x_axis_title=x_title,
                                              x_axis_subtitle=x_subtitle,
                                              y_axis_title=y_title,
                                              y_axis_subtitle=y_subtitle,

                                          ),
                                         )
        for axis in ('x', 'y'):
            for level in (0, 1):
                menus[axis, level].js_on_change('value', menu_callback)
    else:
        menu_callback = build_js_callback(__file__, 'scatter_menu',
                                          args=dict(
                                              scatter_source=scatter_source,
                                              filtered_source=filtered_source,
                                              histogram_source=histogram_source,
                                              x_menu=x_menu,
                                              y_menu=y_menu,
                                              x_axis=fig.xaxis,
                                              y_axis=fig.yaxis,
                                          ),
                                         )
        for menu in [x_menu, y_menu]:
            menu.js_on_change('value', menu_callback)

    format_kwargs = dict(log_scale=bool_to_js(log_scale),
                         identical_bins=bool_to_js(identical_bins),
                        )
    zoom_callback = build_js_callback(__file__, 'scatter_zoom_to_data',
                                      args=dict(
                                          scatter_source=scatter_source,
                                          histogram_source=histogram_source,
                                          x_range=fig.x_range,
                                          y_range=fig.y_range,
                                          hist_x_range=hist_figs['x'].y_range,
                                          hist_y_range=hist_figs['y'].x_range,
                                      ),
                                      format_kwargs=format_kwargs,
                                     )
    zoom_to_data_button.js_on_click(zoom_callback)

    # Toggle labels on points.
    label_button_callback = build_js_callback(__file__, 'scatter_label_button',
                                              args=dict(
                                                  labels=labels,
                                                  label_button=label_button,
                                              ),
                                             )
    label_button.js_on_click(label_button_callback)

    # Set point alpha.
    alpha_callback = build_js_callback(__file__, 'scatter_alpha',
                                       args=dict(
                                           scatter=scatter,
                                       ) 
                                      )
    alpha_widget.js_on_change('value', alpha_callback)


    # Update selection on text search or case sensitive toggle.
    columns_to_search = [c for c in object_cols if c not in color_options]
    search_callback = build_js_callback(__file__, 'scatter_search',
                                        args=dict(
                                            scatter_source=scatter_source,
                                            search_input=search_input,
                                            case_sensitive=case_sensitive,
                                        ),
                                        format_kwargs=dict(column_names=str(columns_to_search)),
                                       )
    search_input.js_on_change('value', search_callback)
    case_sensitive.js_on_change('active', search_callback)

    grid_callback = build_js_callback(__file__, 'scatter_grid',
                                      args=dict(
                                          fig_grid=fig.grid,
                                          fig_diagonal=diagonal,
                                          fig_axes_line=axes_lines,
                                      ),
                                     )
    grid_options.js_on_change('active', grid_callback)

    subset_callback = build_js_callback(__file__, 'scatter_subset_menu',
                                        format_kwargs=dict(subset_indices=str(subset_indices)),
                                        args=dict(
                                            scatter_source=scatter_source,
                                        )
                                       )
    subset_menu.js_on_change('value', subset_callback)

    # Marker size.
    if size_widget_type == 'slider':
        size_widget.js_link('value', scatter.glyph, 'size')
    else:
        size_callback = build_js_callback(__file__, 'scatter_size_menu',
                                          args=dict(
                                              scatter_source=scatter_source,
                                          ),
                                         )
        size_widget.js_on_change('value', size_callback)

    # Save button.
    save_callback = build_js_callback(__file__, 'scatter_save_button',
                                      format_kwargs=dict(column_names=str(table_col_names)),
                                      args=dict(
                                          filtered_source=filtered_source,
                                      ),
                                     )
    save_button.js_on_click(save_callback)

    # Label menu.
    label_callback = build_js_callback(__file__, 'scatter_label',
                                       args=dict(
                                           scatter_source=scatter_source,
                                           filtered_source=filtered_source,
                                       ),
                                      )
    label_menu.js_on_change('value', label_callback)

    # Color menu.
    color_callback = build_js_callback(__file__, 'scatter_color',
                                       args=dict(
                                           scatter_source=scatter_source,
                                       ),
                                      )
    color_menu.js_on_change('value', color_callback)

    # Save annotations.
    annotation_callback = build_js_callback(__file__, 'scatter_annotation_button',
                                            args=dict(
                                                rect_source=rect_source,
                                            ),
                                            format_kwargs=dict(
                                                title=title,
                                            ),
                                           )
    annotation_button.js_on_click(annotation_callback)
    
    if save_as is not None:
        bokeh.io.save(full_layout)
    else:
        bokeh.io.show(full_layout)

    if return_layout:
        return full_layout

def hex_to_CSS(hex_string, alpha=1.):
    ''' Converts an RGB hex value and optional alpha value to a CSS-format RGBA string. '''
    rgb = matplotlib.colors.colorConverter.to_rgb(hex_string)
    rgb = [int(v * 255) for v in rgb]
    CSS = 'rgba({1}, {2}, {3}, {0})'.format(alpha, *rgb)
    return CSS

def parallel_coordinates(df=None,
                         link_axes=True,
                         log_scale=False,
                         save_as=None,
                         initial_limits=None,
                         y_axis_label='',
                         min_start='-Infinity',
                         max_end='Infinity',
                         table_width=1200,
                         width=1200,
                         height=800,
                         label_by=None,
                         extra_injections=None,
                        ):
    ''' Makes an interactive parallel coordinates plot using d3. Call without
    any arguments for an example using data from Jan et al. Science 2014.
    Uses the parallel-coordinates library (github.com/syntagmatic/parallel-coordinates)
    Slickgrid integration taken from github.com/syntagmatic/parallel-coordinates/blob/master/examples/slickgrid.html
    Highlight-on-hover taken from bl.ocks.org/mostaphaRoudsari/b4e090bb50146d88aec4.

    Args:
        df: A pandas DataFrame with columns containing numerical data to plot. 
            Any text columns will be searchable through the 'Search:' field.
            A column named 'color' can be provided to color lines using any
            format understood by matplotlib.colors.to_rgb.
            If df is None, loads example data from Jan et al. Science 2014.

        link_axes: If True, all axes have the same range.

        log_scale: If link_axes=True, whether or not to use a (base 10) log scale.

        initial_limits: If link_axes=True, the intial start and end points of 
            axes limits.

        save_as: A filename (relative to the current directory, and probably 
            ending in .html) in which to save a standalone HTML file containing
            the figure. Opens a download dialog to save the file locally as well.
    '''
    if df is None:
        example_fn = Path(__file__).parent /  'jan_ratios.csv'
        df = pd.read_csv(example_fn, index_col='systematic_name')
        red = '#e41a1c'
        blue = '#0acbee'
        color = pd.Series('black', index=df.index)
        color[df['secretome']] = blue
        color[df['mitop2']] = red
        df['color'] = color

        log_scale = True

    if extra_injections is None:
        extra_injections = {}
    
    # Drop NaNs and copy before changing
    df = df.dropna().copy()
    
    # Collapse multiindex if present
    df.columns = [' '.join(map(str, n)) if isinstance(n, tuple) else str(n) for n in df.columns]

    # Collapsing multindex may mangle color column name.
    color_columns = [c for c in df.columns if c.startswith('color')]

    if len(color_columns) != 1:
        color_series = pd.Series('black', index=df.index)
    else:
        color_column = color_columns[0]
        color_series = df[color_column]
        df.drop(columns=color_column, inplace=True)

    def make_color_string(c):
        array = map(int, np.array(matplotlib.colors.to_rgb(c)) * 255)
        return 'rgba({0}, {1}, {2}, '.format(*array)
    df['_color'] = color_series.map(make_color_string)

    template_fn = Path(__file__).parent / 'template_inline.html'
    html_template = open(template_fn).read()

    encoded_data = base64.b64encode(df.to_csv().encode())
    URI = "'data:text/plain;base64,{0}'".format(encoded_data.decode())

    # Has to be checked in this order since bool subclasses int.
    if isinstance(log_scale, bool):
        log_scale = bool_to_js(log_scale)
    elif isinstance(log_scale, int):
        log_scale = str(log_scale)
    else:
        raise ValueError('log_scale should be an int or a bool', log_scale)

    # template_inline.html contains strings with format /*INJECT:KEY*/ at which data
    # corresponding to values of keys in injections dictionary below will be 
    # substituted into the html, replacing the entire comment-flanked string.
    injections = {
        'encoded_data': URI,
        'link_axes': bool_to_js(link_axes),
        'log_scale': log_scale,
        'initial_limits': str(list(initial_limits)) if initial_limits is not None else 'false',
        'y_axis_label': y_axis_label,
        'min_start': str(min_start),
        'max_end': str(max_end),
        'table_width': str(table_width),
        'width': str(width),
        'height': str(height),
        'label_by': str(label_by),
    }

    injections.update(extra_injections)

    def match_to_injection(match):
        # To accomodate optional injections, return the whole match if not registered.
        return injections.get(match.group(1), match.group(0))

    template_with_data = re.sub("\/\*INJECT:(.*)\*\/", match_to_injection, html_template)
    # Could probably find both /* and <-- in the same regex...
    template_with_data = re.sub("<!--*INJECT:(.*)-->", match_to_injection, template_with_data)
    
    if save_as is not None:
        with open(save_as, 'w') as fh:
            fh.write(template_with_data)
        output = None
    else:
        output = IPython.display.HTML(template_with_data)

    return output

def heatmap(df, cmap, vmin, vmax, height=500):
    normed = matplotlib.colors.Normalize(vmin=vmin, vmax=vmax)(df)
    rgba_float = cmap(normed)
    rgba_int = (rgba_float * 255).astype(np.uint8)[::-1] # flip row order for plotting

    # cryptic numpy manipulations convert from an NxNx4 array of uint8's
    # into an NxN array of uint32's.
    img = np.ascontiguousarray(rgba_int).view(dtype=np.uint32).reshape(df.shape)
    
    num_rows, num_cols = df.shape
    
    width = int(np.floor(height * num_cols / num_rows))
    fig = bokeh.plotting.figure(height=height, width=width, active_scroll='wheel_zoom')
    
    lower_bound = -0.5
    x_upper_bound = lower_bound + num_cols
    
    lower_bound = -0.5
    y_upper_bound = lower_bound + num_rows

    fig.image_rgba(image=[img], x=lower_bound, y=lower_bound, dw=num_cols, dh=num_rows)

    for axis in [fig.yaxis, fig.xaxis]:
        axis.ticker = np.arange(len(df))
        axis.major_label_text_font_size = '0pt'
        
    col_labels = [str(c) for c in df.columns.values]
    row_labels = [str(r) for r in df.index.values]

    fig.xaxis.major_label_overrides = {i: str(c) for i, c in enumerate(col_labels)}
    fig.yaxis.major_label_overrides = {i: l for i, l in enumerate(row_labels[::-1])}

    fig.xaxis.major_label_orientation = 'vertical'

    fig.x_range = bokeh.models.Range1d(lower_bound, x_upper_bound, name='x_range', tags=[x_upper_bound - lower_bound])
    fig.y_range = bokeh.models.Range1d(lower_bound, y_upper_bound, name='y_range')

    fig.grid.visible = False

    fig.axis.major_tick_line_color = None
    fig.axis.major_tick_out = 0
    fig.axis.major_tick_in = 0
    
    quad_source = bokeh.models.ColumnDataSource(data={
        'left': [],
        'right': [],
        'bottom': [],
        'top': [],
    }, name='quad_source',
    )
    fig.quad('left', 'right', 'top', 'bottom',
             fill_alpha=0,
             line_color='black',
             source=quad_source,
            )
    
    search_input = bokeh.models.TextInput(title='Search:', name='search')

    label_fig_size = 200
    label_figs = {
        'left': bokeh.plotting.figure(width=label_fig_size, height=height, x_range=(-10, 10), y_range=fig.y_range),
        'right': bokeh.plotting.figure(width=label_fig_size, height=height, x_range=(-10, 10), y_range=fig.y_range),
        'top': bokeh.plotting.figure(height=label_fig_size, width=width, y_range=(-10, 10), x_range=fig.x_range),
        'bottom': bokeh.plotting.figure(height=label_fig_size, width=width, y_range=(-10, 10), x_range=fig.x_range),
    }

    fig.min_border = 0
    fig.outline_line_color = None
    fig.axis.axis_line_color = None

    for label_fig in label_figs.values():
        label_fig.outline_line_color = None
        label_fig.axis.visible = False
        label_fig.grid.visible = False
        label_fig.min_border = 0
        label_fig.toolbar_location = None
    
    row_label_source = bokeh.models.ColumnDataSource(data={
        'text': row_labels,
        'smallest': [0]*len(row_labels),
        'biggest': [label_fig_size]*len(row_labels),
        'ascending': np.arange(len(row_labels)),
        'descending': np.arange(len(row_labels))[::-1],
        'alpha': [1]*len(row_labels),
    }, name='row_label_source')
    
    col_label_source = bokeh.models.ColumnDataSource(data={
        'text': col_labels,
        'smallest': [0]*len(col_labels),
        'biggest': [label_fig_size]*len(col_labels),
        'ascending': np.arange(len(col_labels)),
        'descending': np.arange(len(col_labels))[::-1],
        'alpha': [1]*len(col_labels),
    }, name='col_label_source')

    common_kwargs = {
        'text': 'text',
        'text_alpha': 'alpha',
        'text_font_size': f'{3000 // (max(x_upper_bound, y_upper_bound) - lower_bound)}%',
        'text_baseline': 'middle',
        'name': 'labels',
    }
    row_kwargs = {
        'y': 'descending',
        'x_units': 'screen',
        'y_units': 'data',
        'source': row_label_source, 
        **common_kwargs,
    }
    col_kwargs = {
        'x': 'ascending',
        'x_units': 'data',
        'y_units': 'screen',
        'source': col_label_source, 
        **common_kwargs,
    }

    labels = {
        'left': bokeh.models.LabelSet(x='biggest',
                                      text_align='right',
                                      x_offset=-10,
                                      **row_kwargs,
                                     ),
        'right': bokeh.models.LabelSet(x='smallest',
                                       text_align='left',
                                       x_offset=10,
                                       **row_kwargs,
                                      ),
        'top': bokeh.models.LabelSet(y='smallest',
                                     text_align='left',
                                     angle=np.pi / 2,
                                     y_offset=10,
                                     **col_kwargs,
                                    ),
        'bottom': bokeh.models.LabelSet(y='biggest',
                                        text_align='right',
                                        angle=np.pi / 2,
                                        y_offset=-10,
                                        **col_kwargs,
                                       ),
    }

    for which in labels:
        label_figs[which].add_layout(labels[which])
    
    fig.toolbar_location = None
    toolbar = bokeh.models.ToolbarBox(toolbar=fig.toolbar)

    # Build callbacks last so that all args are available.

    range_kwargs = dict(lower_bound=lower_bound, upper_bound=x_upper_bound)

    # Prevent ranges from going outside bounds, and resize fonts on zoom.
    range_callback = build_js_callback(__file__, 'heatmap_range',
                                       args=dict(
                                           labels=labels,
                                       ),
                                       format_kwargs=range_kwargs,
                                      )
    for range_ in [fig.x_range, fig.y_range]:
        for prop in ['start', 'end']:
            range_.js_on_change(prop, range_callback)

    search_format_kwargs = dict(
        row_labels=row_labels,
        col_labels=col_labels,
        lower_bound=lower_bound,
        x_upper_bound=x_upper_bound,
        y_upper_bound=y_upper_bound,
    )
    search_callback = build_js_callback(__file__, 'heatmap_search',
                                        args=dict(
                                            quad_source=quad_source,
                                            row_label_source=row_label_source,
                                            col_label_source=col_label_source,
                                        ),
                                        format_kwargs=search_format_kwargs,
                                       )
    search_input.js_on_change('value', search_callback)

    rows = [
        [bokeh.layouts.Spacer(width=label_fig_size), label_figs['top'], bokeh.layouts.Spacer(width=label_fig_size)],
        [label_figs['left'], fig, label_figs['right'], toolbar, search_input],
        [bokeh.layouts.Spacer(width=label_fig_size), label_figs['bottom']],
    ]

    layout = bokeh.layouts.column([bokeh.layouts.row(row) for row in rows])

    bokeh.io.show(layout)



import os
import shlex
import shutil
import subprocess
import tempfile
import threading

from pathlib import Path

import numpy as np
import pysam

from . import genomes, fastq

def build_bowtie2_index(index_prefix, sequence_file_names):
    bowtie2_build_command = ['bowtie2-build',
                             ','.join(map(str, sequence_file_names)),
                             str(index_prefix),
                            ]
    subprocess.check_call(bowtie2_build_command)

class DoNothing(object):
    def __enter__(self):
        pass

    def __exit__(self, exception_type, exception_value, exception_traceback):
        pass

class TemporaryFifo(object):
    def __init__(self, name='fifo'):
        self.name = name

    def __enter__(self):
        self.temp_dir = tempfile.mkdtemp()
        self.file_name = '{0}/{1}'.format(self.temp_dir, self.name)
        os.mkfifo(self.file_name)

    def __exit__(self, exception_type, exception_value, exception_traceback):
        os.remove(self.file_name)
        os.rmdir(self.temp_dir)

class PairedTemporaryFifos(object):
    def __init__(self, name='fifo'):
        self.name = name

    def __enter__(self):
        self.temp_dir = tempfile.mkdtemp()

        # For peace of mind, make sure there are no %'s in the directory name
        # because bowtie2 will be doing a string replacement on %'s.
        while '%' in self.temp_dir:
            os.rmdir(self.temp_dir)
            self.temp_dir = tempfile.mkdtemp()

        self.file_name_template = '{0}/R%_{1}.fastq'.format(self.temp_dir, self.name)

        self.R1_file_name = self.file_name_template.replace('%', '1')
        os.mkfifo(self.R1_file_name)

        self.R2_file_name = self.file_name_template.replace('%', '2')
        os.mkfifo(self.R2_file_name)

    def __exit__(self, exception_type, exception_value, exception_traceback):
        os.remove(self.R1_file_name)
        os.remove(self.R2_file_name)
        os.rmdir(self.temp_dir)

class ThreadFastqWriter(threading.Thread):
    def __init__(self, reads, file_name):
        threading.Thread.__init__(self)
        self.daemon = True
        self.reads = reads
        self.file_name = file_name
        self.start()

    def run(self):
        try:
            with open(self.file_name, 'w') as fifo_fh:
                for i, read in enumerate(self.reads):
                    fifo_fh.write(str(read))

        except BrokenPipeError:
            print('BrokenPipeError caught')

class ThreadPairedFastqWriter(threading.Thread):
    def __init__(self, read_pairs, R1_fn, R2_fn):
        threading.Thread.__init__(self)
        self.daemon = True
        self.read_pairs = read_pairs
        self.R1_fn = R1_fn
        self.R2_fn = R2_fn
        self.start()

    def run(self):
        with open(self.R1_fn, 'w') as R1_fh, open(self.R2_fn, 'w') as R2_fh:
            for R1, R2 in self.read_pairs:
                R1_fh.write(str(R1))
                R2_fh.write(str(R2))

def confirm_index_exists(index_prefix):
    with open(os.devnull, 'w') as devnull:
        try:
            command = ['bowtie2-inspect', '--names', str(index_prefix)]
            subprocess.check_call(command, stdout=devnull, stderr=devnull)
        except subprocess.CalledProcessError:
            raise ValueError(f'Index prefix {index_prefix} does not exist')

def launch_bowtie2(index_prefix,
                   R1_fn,
                   R2_fn,
                   output_file_name,
                   error_file_name,
                   bam_output,
                   sort_by,
                   custom_binary,
                   **options):
    kwarg_to_bowtie2_argument = [
        ('aligned_reads_file_name',   ['--al', options.get('aligned_reads_file_name')]),
        ('unaligned_reads_file_name', ['--un', options.get('unaligned_reads_file_name')]),
        ('aligned_pairs_file_name',   ['--al-conc', options.get('aligned_pairs_file_name')]),
        ('unaligned_pairs_file_name', ['--un-conc', options.get('unaligned_pairs_file_name')]),
        ('omit_secondary_seq',        ['--omit-sec-seq']),
        ('memory_mapped_IO',          ['--mm']),
        ('local',                     ['--local']),
        ('ignore_quals',              ['--ignore-quals']),
        ('threads',                   ['--threads', str(options.get('threads')), '--reorder']),
        ('seed_length',               ['-L', str(options.get('seed_length'))]),
        ('seed_failures',             ['-D', str(options.get('seed_failures'))]),
        ('reseed',                    ['-R', str(options.get('reseed'))]),
        ('seed_mismatches',           ['-N', str(options.get('seed_mismatches'))]),
        ('seed_interval_function',    ['-i', options.get('seed_interval_function')]),
        ('gbar',                      ['--gbar', str(options.get('gbar'))]),
        ('rdg',                       ['--rdg', str(options.get('rdg'))]),
        ('insertion_penalties',       ['--rfg', '{},{}'.format(*options.get('insertion_penalties', (5, 3)))]),
        ('report_all',                ['-a']),
        ('report_up_to',              ['-k', str(options.get('report_up_to'))]),
        ('fasta_input',               ['-f']),
        ('report_timing',             ['-t']),
        ('omit_unaligned',            ['--no-unal']),
        ('min_insert_size',           ['-I', str(options.get('min_insert_size'))]),
        ('max_insert_size',           ['-X', str(options.get('max_insert_size'))]),
        ('forward_forward',           ['--ff']),
        ('score_min',                 ['--score-min', options.get('score_min')]),
        ('maximum_ambiguous',         ['--n-ceil', options.get('maximum_ambiguous')]),
        ('ambiguous_penalty',         ['--np', str(options.get('ambiguous_penalty'))]),
        ('allow_dovetail',            ['--dovetail']),
        ('no_mixed',                  ['--no-mixed']),
        ('num_reads',                 ['--qupto', str(options.get('num_reads'))]),
        ('very_sensitive_local',      ['--very-sensitive-local']),
    ]

    if custom_binary:
        bowtie2_command = [os.environ['HOME'] + '/.local/src/bowtie2-dev/bowtie2']
    else:
        bowtie2_command = ['bowtie2']

    confirm_index_exists(index_prefix)

    for kwarg, bowtie2_argument in kwarg_to_bowtie2_argument:
        if kwarg in options:
            value = options.pop(kwarg)
            # If set to false, don't add the argument.
            if value:
                bowtie2_command.extend(bowtie2_argument)

    # Options are getting popped, so if anything is left, then it was
    # something that wasn't being looked for.
    if len(options) > 0:
        raise ValueError('Unknown keyword argument', options)

    bowtie2_command.extend(['-x', str(index_prefix)])

    if R2_fn != None:
        bowtie2_command.extend(['-1', str(R1_fn)])
        bowtie2_command.extend(['-2', str(R2_fn)])
    else:
        bowtie2_command.extend(['-U', str(R1_fn)])

    error_file = open(error_file_name, 'w')

    if not bam_output:
        bowtie2_command.extend(['-S', output_file_name])

        bowtie2_process = subprocess.Popen(bowtie2_command,
                                           stderr=error_file,
                                          )
        process_to_return = bowtie2_process

    else:
        bowtie2_process = subprocess.Popen(bowtie2_command,
                                           stdout=subprocess.PIPE,
                                           stderr=error_file,
                                          )
        if sort_by is not None:
            sort_command = [
                'samtools', 'sort',
                '-T', output_file_name,
                '-o', output_file_name,
            ]
            if sort_by == 'name':
                sort_command.append('-n')

            sort_command.append('-')

            sort_process = subprocess.Popen(sort_command,
                                            stdin=bowtie2_process.stdout,
                                            stderr=subprocess.PIPE,
                                           )
            bowtie2_process.stdout.close()
            process_to_return = sort_process

        else:
            view_command = [
                'samtools', 'view', '-b',
                '-o', output_file_name,
            ]
            view_process = subprocess.Popen(view_command,
                                            stdin=bowtie2_process.stdout,
                                            stderr=subprocess.PIPE,
                                           )
            bowtie2_process.stdout.close()
            process_to_return = view_process

    return process_to_return, bowtie2_command

def _map_bowtie2(index_prefix,
                 R1_fn,
                 R2_fn,
                 output_file_name,
                 error_file_name='/dev/null',
                 custom_binary=False,
                 bam_output=False,
                 sort_by=None,
                 reads=None,
                 read_pairs=None,
                 yield_mappings=False,
                 yield_unaligned=False,
                 **options):

    using_input_fifos = reads is not None or read_pairs is not None
    is_paired = R2_fn is not None or read_pairs is not None

    if reads is not None:
        input_fifo_source = TemporaryFifo(name='input_fifo.fastq')
    elif read_pairs is not None:
        input_fifo_source = PairedTemporaryFifos(name='input')
    else:
        input_fifo_source = DoNothing()

    if yield_unaligned:
        if is_paired:
            output_fifo_source = PairedTemporaryFifos(name='unaligned')
        else:
            output_fifo_source = TemporaryFifo(name='unaligned.fastq')
    elif yield_mappings:
        output_fifo_source = TemporaryFifo(name='output_fifo.sam')
    else:
        output_fifo_source = DoNothing()

    with input_fifo_source, output_fifo_source:
        if reads is not None:
            R1_fn = input_fifo_source.file_name
            writer = ThreadFastqWriter(reads, R1_fn)
        elif read_pairs is not None:
            R1_fn = input_fifo_source.R1_file_name
            R2_fn = input_fifo_source.R2_file_name
            writer = ThreadPairedFastqWriter(read_pairs, R1_fn, R2_fn)

        if yield_unaligned:
            if is_paired:
                unal_template_fn = output_fifo_source.file_name_template
                unal_R1_fn = output_fifo_source.R1_file_name
                unal_R2_fn = output_fifo_source.R2_file_name
                options['unaligned_pairs_file_name'] = unal_template_fn
            else:
                unal_R1_fn = output_fifo_source.file_name
                options['unaligned_reads_file_name'] = unal_R1_fn
        elif yield_mappings:
            output_file_name = output_fifo_source.file_name

        bowtie2_process, bowtie2_command = launch_bowtie2(index_prefix,
                                                          R1_fn,
                                                          R2_fn,
                                                          output_file_name,
                                                          error_file_name,
                                                          bam_output,
                                                          sort_by,
                                                          custom_binary,
                                                          **options)

        if yield_unaligned:
            if is_paired:
                for read_pair in fastq.read_pairs(unal_R1_fn, unal_R2_fn):
                    yield read_pair
            else:
                for read in fastq.reads(unal_R1_fn):
                    yield read

        elif yield_mappings:
            sam_file = pysam.AlignmentFile(str(output_file_name), 'r')
            yield sam_file
            for read in sam_file:
                yield read

        _, err_output = bowtie2_process.communicate()
        if bowtie2_process.returncode != 0:
            raise subprocess.CalledProcessError(bowtie2_process.returncode,
                                                bowtie2_command,
                                                err_output,
                                               )
        if bam_output and sort_by == 'position':
            pysam.index(str(output_file_name))

def map_bowtie2(index_prefix,
                R1_fn=None,
                R2_fn=None,
                output_file_name=None,
                bam_output=False,
                sort_by=None,
                error_file_name='/dev/null',
                custom_binary=False,
                reads=None,
                read_pairs=None,
                yield_mappings=False,
                yield_unaligned=False,
                **options):

    if reads is not None and read_pairs is not None:
        raise RuntimeError('Can\'t give unpaired_Reads and paired_Reads')

    if yield_unaligned and yield_mappings:
        raise RuntimeError('Can\'t yield unaligned and mappings.')

    if yield_mappings and bam_output:
        raise RuntimeError('yield_mappings and bam_output can\'t both be True.')

    if output_file_name == None and yield_mappings == False:
        raise RuntimeError('Need to give output_file_name or yield_mappings')

    if read_pairs:
        # Can't used named pipes for paired Reads without custom binary because
        # of buffer size mismatch.
        custom_binary = True

    generator = _map_bowtie2(index_prefix,
                             R1_fn=R1_fn,
                             R2_fn=R2_fn,
                             output_file_name=str(output_file_name),
                             error_file_name=error_file_name,
                             custom_binary=custom_binary,
                             bam_output=bam_output,
                             sort_by=sort_by,
                             reads=reads,
                             read_pairs=read_pairs,
                             yield_mappings=yield_mappings,
                             yield_unaligned=yield_unaligned,
                             **options)
    if yield_unaligned:
        return generator

    elif yield_mappings:
        sam_file = next(generator)
        return sam_file, generator

    else:
        # There isn't a real yield in generator, so calling next() is just
        # executing the function.
        try:
            next(generator)
        except StopIteration:
            pass

def map_tophat(reads_file_names,
               bowtie2_index,
               tophat_dir,
               gtf_file_name=None,
               transcriptome_index=None,
               num_threads=1,
               no_sort=False,
               keep_temporary_files=True,
              ):

    joined_reads_names = ','.join(str(fn) for fn in reads_file_names)

    options = [
        '--no-novel-juncs',
        '--num-threads', str(num_threads),
        '--output-dir', str(tophat_dir),
        '--report-secondary-alignments',
        '--read-realign-edit-dist', '0',
        #'--read-mismatches', '6',
        #'--read-edit-dist', '6',
        #'--b2-n-ceil', 'C,10,0',
        #'--b2-L', '10',
    ]

    if gtf_file_name is not None:
        options.extend(['--GTF', str(gtf_file_name)])
    if transcriptome_index is not None:
        options.extend(['--transcriptome-index', str(transcriptome_index)])
    if no_sort:
        options.append('--no-sort-bam')
    if keep_temporary_files:
        options.append('--keep-tmp')

    tophat_command = ['tophat2'] + options + [str(bowtie2_index), joined_reads_names]

    # tophat maintains its own logs of everything that is written to the
    # console, so discard output.
    try:
        output = subprocess.check_output(tophat_command, stderr=subprocess.STDOUT)
    except subprocess.CalledProcessError as e:
        print(f'tophat command returned code {e.returncode}')
        print(f'full command was:\n\t{shlex.join(tophat_command)}')
        print(f'output from tophat was:\n\t{e.output}')
        raise ValueError

    # If there were no unmapped reads, tophat won't create a file. I want to be
    # able to assume that one exists.
    accepted_hits_fn = f'{tophat_dir}/accepted_hits.bam'
    unmapped_fn = f'{tophat_dir}/unmapped.bam'
    if not os.path.exists(unmapped_fn):
        template = pysam.AlignmentFile(accepted_hits_fn)
        empty_unmapped = pysam.AlignmentFile(unmapped_fn, 'wb', template=template)
        template.close()
        empty_unmapped.close()

    if not no_sort:
        pysam.index(accepted_hits_fn)

def map_tophat_paired(R1_fn,
                      R2_fn,
                      bowtie2_index,
                      gtf_file_name,
                      transcriptome_index,
                      tophat_dir,
                      num_threads=1,
                      no_sort=False,
                     ):
    options = [
        '--GTF', gtf_file_name,
        '--no-novel-juncs',
        '--num-threads', str(num_threads),
        '--output-dir', tophat_dir,
        '--transcriptome-index', transcriptome_index,
        '--report-secondary-alignments',
        '--read-realign-edit-dist', '0',
    ]
    if no_sort:
        options.append('--no-sort-bam')

    tophat_command = ['tophat2'] + options + [bowtie2_index, R1_fn, R2_fn]

    # tophat maintains its own logs of everything that is written to the
    # console, so discard output.
    subprocess.check_output(tophat_command, stderr=subprocess.STDOUT)

    accepted_hits_fn = '{0}/accepted_hits.bam'.format(tophat_dir)
    if not no_sort:
        pysam.index(accepted_hits_fn)

def run_STAR_command(STAR_command, clean_up_cwd=False):
    try:
        subprocess.run(STAR_command,
                       check=True,
                       stdout=subprocess.PIPE,
                       stderr=subprocess.STDOUT,
                      )

        if clean_up_cwd:
            clean_up_STAR_output(f'{os.getcwd()}{os.sep}')

    except subprocess.CalledProcessError as e:
        print(f'STAR command returned code {e.returncode}')
        print(f'full command was:\n\n{shlex.join(STAR_command)}')
        print(f'output from STAR was:\n\n{e.output.decode()}\n')
        raise

def map_STAR(R1_fn, index_dir, output_prefix,
             R2_fn=None,
             num_threads=1,
             num_reads=-1,
             sort=True,
             include_unmapped=False,
             bam_fn=None,
             mode='stringent',
             clean_up_afterwards=True,
             use_shared_memeory=True,
            ):
    if sort:
        bam_suffix = 'Aligned.sortedByCoord.out.bam'
        sort_option = 'SortedByCoordinate'
    else:
        bam_suffix = 'Aligned.out.bam'
        sort_option = 'Unsorted'

    if include_unmapped:
        unmapped_option = 'Within'
    else:
        unmapped_option = 'None'

    def make_string(possibly_list):
        if isinstance(possibly_list, list):
            string = ','.join(map(str, possibly_list))
            first_fn = Path(possibly_list[0])
        else:
            string = str(possibly_list)
            first_fn = Path(possibly_list)

        return string, first_fn.suffix

    R1_fn_string, R1_suffix = make_string(R1_fn)

    STAR_command = [
        'STAR',
        '--genomeDir', str(index_dir),
        '--outSAMtype', 'BAM', sort_option,
        '--outSAMunmapped', unmapped_option,
        '--outSAMattributes', 'All',
        '--limitBAMsortRAM', '1345513406',
        '--alignIntronMax', '1',
        '--runThreadN', str(num_threads),
        '--readMapNumber', str(num_reads),
        '--outFileNamePrefix', str(output_prefix),
    ]

    if use_shared_memeory:
        STAR_command.extend([
        '--genomeLoad', 'LoadAndKeep',
    ])

    if mode == 'stringent':
        STAR_command.extend([
            '--outFilterScoreMinOverLread', '0.2',
            '--outFilterMatchNminOverLread', '0.2',
            '--outFilterMatchNmin', '50',
        ])

    elif mode == 'permissive':
        STAR_command.extend([
            '--outFilterMultimapScoreRange', '1000',
            '--outFilterMultimapNmax', '1000',
            '--outFilterScoreMinOverLread', '0',
            '--outFilterMatchNminOverLread', '0',
            '--outFilterMatchNmin', '10',
        ])

    elif mode == 'guide_alignment':
        STAR_command.extend([
            '--alignEndsType', 'EndToEnd',
            '--outFilterMultimapNmax', '1000',
        ])

    else:
        pass

    STAR_command.extend([
        '--readFilesIn', R1_fn_string,
    ])

    if R2_fn is not None:
        R2_fn_string, R2_suffix = make_string(R2_fn)
        STAR_command.append(R2_fn_string)

    if R1_suffix == '.gz':
        STAR_command.extend(['--readFilesCommand', 'zcat'])
    elif R1_suffix == '.bam':
        STAR_command.extend(['--readFilesCommand', 'samtools view',
                             '--readFilesType', 'SAM SE',
                            ])

    run_STAR_command(STAR_command)

    initial_bam_fn = str(output_prefix) + bam_suffix

    if bam_fn is None:
        bam_fn = initial_bam_fn
    else:
        shutil.move(str(initial_bam_fn), str(bam_fn))

    if sort:
        pysam.index(str(bam_fn))

    if clean_up_afterwards:
        clean_up_STAR_output(output_prefix)

    return bam_fn

def load_STAR_index(index_dir):
    STAR_command = [
        'STAR',
        '--genomeDir', str(index_dir),
        '--genomeLoad', 'LoadAndExit',
    ]
    run_STAR_command(STAR_command, clean_up_cwd=True)

def remove_STAR_index(index_dir):
    STAR_command = [
        'STAR',
        '--genomeDir', str(index_dir),
        '--genomeLoad', 'Remove',
    ]
    run_STAR_command(STAR_command, clean_up_cwd=True)

def build_STAR_index(fasta_files, index_dir, wonky_param=None, num_threads=1, RAM_limit=None):
    total_length = 0
    for fasta_file in fasta_files:
        for name, entry in genomes.parse_fai(str(fasta_file) + '.fai').items():
            total_length += entry.length

    if wonky_param is None:
        wonky_param = int(min(14, np.log2(total_length) / 2. - 1))

    STAR_command = [
        'STAR',
        '--runMode', 'genomeGenerate',
        '--genomeDir', str(index_dir),
        '--genomeFastaFiles', *fasta_files,
        '--genomeSAindexNbases', str(wonky_param),
        '--runThreadN', str(num_threads),
    ]

    if RAM_limit is not None:
        STAR_command.extend(['--limitGenomeGenerateRAM', str(RAM_limit)])

    run_STAR_command(STAR_command, clean_up_cwd=True)

def clean_up_STAR_output(output_prefix):
    suffixes_to_clean_up = [
        'Log.final.out',
        'Log.out',
        'Log.progress.out',
        'SJ.out.tab',
        '_STARtmp',
    ]

    for suffix in suffixes_to_clean_up:
        full_fn = Path(str(output_prefix) + suffix)
        if full_fn.exists():
            if full_fn.is_dir():
                shutil.rmtree(full_fn)
            else:
                full_fn.unlink()

def map_minimap2(fastq_fn, index, bam_fn, num_threads=1):
    minimap2_command = [
        'minimap2',
        '-a', # sam output
        '-Y', # use soft clipping for supplementary alignments instead of hard clipping
        '-P', # (roughly equivalent to?) report all
        '--MD', # populate MD tags
        '-r', '20', # max bandwidth
        '-t', str(num_threads),
        str(index),
        str(fastq_fn),
    ]

    minimap2_process = subprocess.Popen(minimap2_command,
                                        stdout=subprocess.PIPE,
                                        stderr=subprocess.PIPE,
                                       )

    view_command = ['samtools', 'view', '-b', '-o', str(bam_fn)]
    view_process = subprocess.Popen(view_command,
                                    stdin=minimap2_process.stdout,
                                    stdout=subprocess.PIPE,
                                   )

    minimap2_process.stdout.close()
    view_out, view_err = view_process.communicate()

    minimap2_process.wait()

    if minimap2_process.returncode != 0:
        raise subprocess.CalledProcessError(minimap2_process.returncode, minimap2_process.args)

def build_minimap2_index(fasta_fn, index_fn):
    minimap2_command = [
        'minimap2',
        '-H',
        '-d', str(index_fn),
        str(fasta_fn),
    ]

    try:
        subprocess.run(minimap2_command,
                       check=True,
                       stdout=subprocess.PIPE,
                       stderr=subprocess.PIPE,
                      )
    except subprocess.CalledProcessError as e:
        print(f'minimap2 command returned code {e.returncode}')
        print(f'full command was:\n\n{shlex.join(minimap2_command)}\n')
        print(f'stdout from minimap2 was:\n\n{e.stdout.decode()}\n')
        print(f'stderr from minimap2 was:\n\n{e.stderr.decode()}\n')
        raise



import numpy as np

extension = 'txt'

def write_file(array, file_name):
    ''' Writes a 2D array into a text file.
    '''
    if array.ndim < 2:
        array = np.atleast_2d(array)
    elif array.ndim == 2:
        np.savetxt(file_name, array, fmt='%i', delimiter='\t')
    else:
        raise RuntimeError('Array has more than 2 dimensions.')

def read_file(file_name):
    ''' Reads the format written by array_file.write into an array.
    '''
    array = np.loadtxt(file_name, dtype=np.int, ndmin=2)
    return array

def combine_data(first_array, second_array):
    ''' Identically shaped arrays are added.
        Arrays of different sizes are treated as the smallest upper-left
        corner of an eventually-zero-in-both-directions infinite array.
    '''
    first_rows, first_cols = first_array.shape
    second_rows, second_cols = second_array.shape
    
    max_rows = max(first_rows, second_rows)
    max_cols = max(first_cols, second_cols)

    combined = np.zeros((max_rows, max_cols), int)
    combined[:first_rows, :first_cols] += first_array
    combined[:second_rows, :second_cols] += second_array
    return combined



import unittest
from collections import defaultdict

from hits import parallel
from hits.parallel import split_file as split_file
from hits import fastq as fastq

class TestSplit(unittest.TestCase):

    def test_piece_of_list(self):
        ''' Tests whether joining all which_pieces from piece_of_list recreates
            the original list.
        '''
        list_sizes = range(1000, 1100)
        for size in list_sizes:
            original_list = list(range(size))
            num_pieces = 96
            pieces = []
            pieces_joined = []
            for which_piece in range(num_pieces):
                piece = parallel.piece_of_list(original_list,
                                               num_pieces,
                                               which_piece,
                                              )
                pieces.append(piece)
                pieces_joined.extend(piece)

            self.assertEqual(
                set(original_list), set(pieces_joined),
                msg='Failed set equality for list size {0}'.format(size),
            )

            piece_lengths = [len(p) for p in pieces]
            self.assertTrue(
                max(piece_lengths) - min(piece_lengths) <= 1,
                msg='Pieces too variable in size for list size {0}'.format(size),
            )

    def test_split_fastq(self):
        ''' Tests whether every read in a fastq file shows up exactly once in
        piece of a split.
        '''
        fn = 'data.fastq'

        whole = defaultdict(list)

        for read in fastq.reads(fn):
            whole[read.name].append(read)

        num_pieces_list = [1, 10, 100]
        from_pieces = {n: defaultdict(list) for n in num_pieces_list}

        for num_pieces in num_pieces_list:
            for which_piece in range(num_pieces):
                piece = split_file.piece(fn, num_pieces, which_piece, 'fastq')

                for read in fastq.reads(piece):
                    from_pieces[num_pieces][read.name].append(read)

            self.assertEqual(
                whole, from_pieces[num_pieces],
                msg='Splitting did not partition',
            )

if __name__ == '__main__':
    unittest.main()



import shutil
from functools import partial, reduce
from .. import external_sort, sam
from . import log

def _concatenate(input_file_names, output_file_name):
    with open(output_file_name, 'w') as output_file:
        for input_file_name in input_file_names:
            shutil.copyfileobj(open(input_file_name), output_file)

def _merge_sam_files(input_file_names, merged_file_name, are_sorted=False):
    ''' Merges a list of sam files.
        Requires all input files to have the same @SQ lines.
    '''
    sq_lines = None
    for file_name in input_file_names:
        these_sq_lines = sam.get_sq_lines(file_name)
        if sq_lines == None:
            sq_lines = these_sq_lines
        else:
            if sq_lines != these_sq_lines:
                raise ValueError('@SQ lines do not agree')

    with open(merged_file_name, 'w') as merged_file:
        for sq_line in sq_lines:
            merged_file.write(sq_line)

        input_files = [sam.open_to_reads(fn) for fn in input_file_names]
        if are_sorted:
            for line in external_sort.merge(input_files):
                merged_file.write(line)
        else:
            for input_file in input_files:
                shutil.copyfileobj(input_file, merged_file)

special_mergers = {
    'bam': sam.merge_sorted_bam_files,
    'bam_by_name': partial(sam.merge_sorted_bam_files, by_name=True),
    'sam_unsorted':  partial(_merge_sam_files, are_sorted=False),
    'sam_sorted': partial(_merge_sam_files, are_sorted=True),
    'fastq': _concatenate,
    'fasta': _concatenate,
    'concatenate': _concatenate,
}

def merge_files(input_file_names, output_file_name, file_format, fast=False):
    try:
        if fast:
            file_format.fast_merge(input_file_names, output_file_name)
        elif isinstance(file_format, str):
            special_mergers[file_format](input_file_names, output_file_name)
        else:
            processed_inputs = (file_format.read_file(fn) for fn in input_file_names)
            merged_data = reduce(file_format.combine_data, processed_inputs)
            file_format.write_file(merged_data, output_file_name)
    except IOError:
        print(output_file_name)
        raise
